# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file tab20_maps.py Maps Tab
##
## This plugin displays a list of all WAD addons that contain maps.
## Selections are equivalent to the Addons tab, but they are presented
## as a list control where it's easier to select one addon instead of
## checking/unchecking a tree.

import os
import ui, events, language
import widgets as wg
import sb.widget.button as wb
import sb.widget.list
import sb.aodb as ao
import sb.profdb as pr
import widgets


MAPS = 'tab-maps'


# React to selection notifications.
listenSelections = True

# The map list box (has columns).
mapListBox = None


def init():
    # Create the Maps page.
    area = ui.createTab(MAPS)
    area.setBorderDirs(ui.BORDER_NOT_BOTTOM)
    area.setWeight(1)

    global mapListBox
    area.setBorderDirs(ui.BORDER_NOT_BOTTOM)
    mapListBox = area.createList('maps-list', sb.widget.list.List.STYLE_COLUMNS)

    # The columns.
    mapListBox.addColumn('maps-list-icon', 20)
    mapListBox.addColumn('maps-list-identifier', 180)
    mapListBox.addColumn('maps-list-count')

    # Prepare icons.
    global iconManager, mapIcons
    iconManager = widgets.IconManager(16, 16)
    mapIcons = [iconManager.get('unchecked'), iconManager.get('checked')]
    mapListBox.setImageList(iconManager.getImageList())

    # Some buttons in the bottom.
    area.setWeight(0)
    area.setBorderDirs(ui.BORDER_NOT_TOP)
    buttonArea = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=4)
    buttonArea.setBorderDirs(ui.BORDER_TOP | ui.BORDER_RIGHT)
    buttonArea.setWeight(0)
    clearButton = buttonArea.createButton('maps-list-clear', wb.Button.STYLE_MINI)
    clearButton.resizeToBestSize()
    refreshButton = buttonArea.createButton('refresh-addon-database', wb.Button.STYLE_MINI)
    refreshButton.resizeToBestSize()   
    
    buttonArea.setWeight(1)
    buttonArea.addSpacer()
    buttonArea.setWeight(0)
    buttonArea.setBorderDirs(ui.BORDER_TOP | ui.BORDER_LEFT)
    buttonArea.createButton('show-addon-paths')   

    # Listen to our commands.
    events.addCommandListener(handleCommand, ['maps-list-clear'])

    events.addNotifyListener(handleNotify, ['active-profile-changed',
                                            'maps-list-selected',
                                            'maps-list-deselected',
                                            'addon-attached',
                                            'addon-detached',
                                            'addon-database-reloaded',
                                            'language-changed'])

    
def handleCommand(event):
    """Handle a command event."""
    
    if event.hasId('maps-list-clear'):
        mapListBox.selectItem(None)


def handleNotify(event):
    """Handle a notification event."""

    global listenSelections

    if event.hasId('active-profile-changed'):
        refreshList()

    if event.hasId('maps-list-selected') and listenSelections:
        pr.getActive().useAddon(event.getSelection())

    if event.hasId('maps-list-deselected'):
        pr.getActive().dontUseAddon(event.getDeselection())

    if event.hasId('addon-attached'):
        listenSelections = False
        mapListBox.selectItem(event.getAddon())
        mapListBox.setItemImage(event.getAddon(), 1)
        listenSelections = True

    if event.hasId('addon-detached'):
        listenSelections = False
        mapListBox.deselectItem(event.getAddon())
        mapListBox.setItemImage(event.getAddon(), 0)
        listenSelections = True
        
    if event.hasId('addon-database-reloaded'):
        refreshList()
        
    if event.hasId('language-changed'):
        mapListBox.retranslateColumns()
    

def refreshList():
    """Fill the maps list with addons."""

    # Clear the list.
    mapListBox.clear()

    wads = [a for a in ao.getAvailableAddons(pr.getActive())
            if a.getType() == 'addon-type-wad' and a.isPWAD()]

    # Translate addon names.
    wads.sort(lambda a, b: cmp(a.getId(), b.getId()))

    for wad in wads:
        # TODO: More information titles.
        if not language.isDefined(wad.getId()):
            visibleName = language.translate(wad.getId())
        else:
            visibleName = os.path.basename(wad.getContentPath())
        mapListBox.addItemWithColumns(wad.getId(),
                                      0,
                                      visibleName,
                                      wad.getShortContentAnalysis())

    prof = pr.getActive()
    usedAddons = prof.getUsedAddons()
    theFirst = True

    # Select the addons currently attached to the profile.
    # Also make sure the first one is visible.
    for addonId in usedAddons:
        mapListBox.selectItem(addonId)
        mapListBox.setItemImage(addonId, 1)
        if theFirst:
            # Make sure it's visible.
            mapListBox.ensureVisible(addonId)
            theFirst = False



#!/usr/bin/env python2.7
#
# Doomsday Build Configuration Script
#
# Usage:
#   configure.py
#
# This will generate versions.pri based on various header files.

import sys, os, time

sys.path.append(os.path.join(os.path.dirname(__file__), '../distrib'))
import build_version

out = file('versions.pri', 'wt')
print >> out, "# Generated by configure.py on", time.asctime()

# Check primary version info.
primary = build_version.version_summary('api/dd_version.h').split(' ')

base = primary[0]
buildNum = primary[1]
relType = primary[2]

print >> out, """
defineReplace(dengReleaseType) {
    return(%s)
}""" % relType

print >> out, """
defineTest(isStableRelease) {
    relType = $$dengReleaseType()
    contains(relType, Stable): return(true)
    else: return(false)
}
"""

print >> out, 'DENG_VERSION =', base
print >> out, 'JDOOM_VERSION =', base
print >> out, 'JHERETIC_VERSION =', base
print >> out, 'JHEXEN_VERSION =', base
print >> out, 'JDOOM64_VERSION =', base

def find_version(fn):
    return build_version.version_summary(fn).split(' ')[0]

def print_version(out, id, fn):
    print >> out, '%s_VERSION = %s' % (id, find_version(fn))

print_version(out, 'DEHREAD',           'plugins/dehread/include/version.h')
print_version(out, 'SAVEGAMECONVERTER', 'plugins/savegameconverter/include/version.h')
print_version(out, 'IDTECH1CONVERTER',  'plugins/idtech1converter/include/version.h')
print_version(out, 'DIRECTSOUND',       'plugins/directsound/include/version.h')
print_version(out, 'OPENAL',            'plugins/openal/include/version.h')
print_version(out, 'FMOD',              'plugins/fmod/include/version.h')
print_version(out, 'WINMM',             'plugins/winmm/include/version.h')
print_version(out, 'EXAMPLE',           'plugins/example/include/version.h')



#!/usr/bin/env python3
# -*- Coding: UTF-8 -*-

# ---------------------------------------------------------------------------
# Open Asset Import Library (ASSIMP)
# ---------------------------------------------------------------------------
#
# Copyright (c) 2006-2010, ASSIMP Development Team
#
# All rights reserved.
#
# Redistribution and use of this software in source and binary forms, 
# with or without modification, are permitted provided that the following 
# conditions are met:
# 
# * Redistributions of source code must retain the above
#   copyright notice, this list of conditions and the
#   following disclaimer.
# 
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the
#   following disclaimer in the documentation and/or other
#   materials provided with the distribution.
# 
# * Neither the name of the ASSIMP team, nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior
#   written permission of the ASSIMP Development Team.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ---------------------------------------------------------------------------

"""Generate the regression database db.zip from the files in the <root>
/test/models directory. Older databases are overwritten with no prompt.
"""

import sys
import os
import subprocess
import zipfile

import settings
import utils

usage = """gen_db [-i=...] [-e=...] [-p] [-n]

(lists of file extensions are comma delimited, i.e. `3ds,lwo,x`)
-i,--include: List of file extensions to update dumps for. If omitted,
         all file extensions are updated except those in `exclude`.

-e,--exclude: Merged with settings.exclude_extensions to produce a
         list of all file extensions to ignore. If dumps exist,
         they are not altered. If not, theu are not created.

-p,--preview: Preview list of file extensions touched by the update.
         Dont' change anything.

-n,--nozip: Don't pack to ZIP archive. Keep all dumps in individual files.
"""

# -------------------------------------------------------------------------------
def process_dir(d, outfile, file_filter):
    """ Generate small dump records for all files in 'd' """
    print("Processing directory " + d)

    num = 0
    for f in os.listdir(d):
        fullp = os.path.join(d, f)
        if os.path.isdir(fullp) and not f == ".svn":
            num += process_dir(fullp, outfile, file_filter)
            continue

        if file_filter(f):
            for pp in settings.pp_configs_to_test:
                num += 1
                print("DUMP " + fullp + "\n post-processing: " + pp)
                outf = os.path.join(os.getcwd(), settings.database_name,
                    utils.hashing(fullp, pp))

                cmd = [utils.assimp_bin_path,"dump",fullp,outf,"-b","-s","-l"] + pp.split()
                outfile.write("assimp dump "+"-"*80+"\n")
                outfile.flush()
                if subprocess.call(cmd, stdout=outfile, stderr=outfile, shell=False):
                    print("Failure processing " + fullp)

                    # spit out an empty file to indicate that this failure is expected
                    with open(outf,'wb') as f:
                        pass
    return num
                    

# -------------------------------------------------------------------------------
def make_zip():
    """Zip the contents of ./<settings.database_name>
    to <settings.database_name>.zip using DEFLATE
    compression to minimize the file size. """

    num = 0
    zipout = zipfile.ZipFile(settings.database_name + ".zip", "w", zipfile.ZIP_DEFLATED)
    for f in os.listdir(settings.database_name):
        p = os.path.join(settings.database_name, f)
        zipout.write(p, f)
        if settings.remove_old:
            os.remove(p)

        num += 1

    if settings.remove_old:
        os.rmdir(settings.database_name)

    bad = zipout.testzip()
    assert bad is None

    print("="*60)
    print("Database contains {0} entries".format(num))


# -------------------------------------------------------------------------------
def extract_zip():
    """Unzip <settings.database_name>.zip to
    ./<settings.database_name>"""
    try:
        zipout = zipfile.ZipFile(settings.database_name + ".zip", "r", 0)
        zipout.extractall(path=settings.database_name)
    except (RuntimeError,IOError) as r:
        print(r)
        print("failed to extract previous ZIP contents. "\
              "DB is generated from scratch.")
        

# -------------------------------------------------------------------------------
def gen_db(ext_list,outfile):
    """Generate the crash dump database in
    ./<settings.database_name>"""
    try:
        os.mkdir(settings.database_name)
    except OSError:
        pass

    num = 0
    for tp in settings.model_directories:
        num += process_dir(tp, outfile,
            lambda x: os.path.splitext(x)[1] in ext_list)

    print("="*60)
    print("Updated {0} entries".format(num))
        

# -------------------------------------------------------------------------------
if __name__ == "__main__":
    utils.find_assimp_or_die()
    def clean(f):
        f = f.strip("* \'")
        return "."+f if f[:1] != '.' else f

    if len(sys.argv)>1 and (sys.argv[1] == "--help" or sys.argv[1] == "-h"):
        print(usage)
        sys.exit(0)

    ext_list, preview, nozip = None, False, False
    for m in sys.argv[1:]:
        if m[:10]=="--exclude=":
            settings.exclude_extensions += map(clean, m[10:].split(","))
        elif m[:3]=="-e=":
            settings.exclude_extensions += map(clean, m[3:].split(","))
        elif m[:10]=="--include=":
            ext_list = m[10:].split(",")
        elif m[:3]=="-i=":
            ext_list = m[3:].split(",")
        elif m=="-p" or m == "--preview":
            preview = True
        elif m=="-n" or m == "--nozip":
            nozip = True
            
    outfile = open(os.path.join("..", "results", "gen_regression_db_output.txt"), "w")
    if ext_list is None:
        (ext_list, err) = subprocess.Popen([utils.assimp_bin_path, "listext"],
            stdout=subprocess.PIPE).communicate()
        ext_list = str(ext_list).lower().split(";")

    # todo: Fix for multi dot extensions like .skeleton.xml
    ext_list = list(filter(lambda f: not f in settings.exclude_extensions,
        map(clean, ext_list)))

    if preview:
        print(','.join(ext_list))
        sys.exit(1)

    extract_zip()    
    gen_db(ext_list,outfile)
    make_zip()

    print("="*60)
    input("Press any key to continue")
    sys.exit(0)
    
# vim: ai ts=4 sts=4 et sw=4    




# -*- coding: iso-8859-1 -*-
# $Id: main.py 4303 2007-01-01 19:06:40Z skyjake $
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file loadorder.py Addon Load Order Dialog
##
## This module implements the load order dialog.

import os, time, string
import paths, events, ui, language
import sb.util.dialog
import sb.widget.button as wg
import sb.widget.list
import sb.profdb as pr
import sb.confdb as st
import sb.aodb as ao
import sb.addon


def run(profile):
    """Open the load order dialog.

    @param profile The profile whose order is being edited.
    """

    dialog, area = sb.util.dialog.createButtonDialog(
        'load-order-dialog',
        ['reset', '', 'cancel', 'ok'], 'ok')

    dialog.focusWidget('cancel')
    area.setWeight(0)
    area.createText('load-order-message')

    # Create the widgets that will be used to edit the order.
    area.setWeight(1)
    top = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)

    top.setWeight(3)
    listArea = top.createArea(border=3)
    listArea.setWeight(1)
    orderList = listArea.createList('')
    orderList.setMinSize(200, 200)

    top.setWeight(1)
    controlArea = top.createArea(border=3)
    controlArea.setWeight(0)
    controlArea.setExpanding(False)
    upButton = controlArea.createButton('load-order-move-up')
    downButton = controlArea.createButton('load-order-move-down')
    controlArea.addSpacer()
    firstButton = controlArea.createButton('load-order-move-top')
    lastButton = controlArea.createButton('load-order-move-bottom')

    def moveSelFirst():
        orderList.moveItem(orderList.getSelectedItem(), -65535)

    def moveSelLast():
        orderList.moveItem(orderList.getSelectedItem(), +65535)

    def moveSelUp():
        orderList.moveItem(orderList.getSelectedItem(), -1)

    def moveSelDown():
        orderList.moveItem(orderList.getSelectedItem(), +1)

    firstButton.addReaction(moveSelFirst)
    upButton.addReaction(moveSelUp)
    downButton.addReaction(moveSelDown)
    lastButton.addReaction(moveSelLast)

    # The final addons are automatically sorted according to the
    # profile's load order.
    finalAddons = profile.getFinalAddons()
    
    # Put all the finalized list of loaded addons into the list.
    for ident in finalAddons:
        orderList.addItem(ident)

    if len(finalAddons):
        orderList.selectItem(finalAddons[0])

    # We'll accept 'reset' as a dialog end command as well.
    dialog.addEndCommand('reset')

    # Run the dialog and handle the closing command.
    result = dialog.run()

    if result == 'ok':
        # This is the new load order for the profile.
        profile.setLoadOrder(orderList.getItems())

    elif result == 'reset':
        # Clear the profile's load order.
        profile.setLoadOrder([])

    else:
        # Nothing is done on cancel.
        pass




# Reads a cphelp/conhelp file and outputs an .ame source for each
# command and variable.
#
# Arguments:
# 1 - input file (cphelp/conhelp)
# 2 - component (engine, libcommon, etc.)

import os
import sys

def escaped(s):
    result = ''
    for c in s:
        if c == '{': c = '@{'
        elif c == '}': c = '@}'
        elif c == '@': c = '@@'
        result += c
    return result

source = sys.argv[1]
component = sys.argv[2]
mode = None
name = ''
path = None
output = None

print 'Component:', component

for line in file(source).readlines():
    line = line.strip()
    if len(line) == 0: continue
    
    if '# CONSOLE VARIABLE' in line: mode = 'variable'
    elif '# CONSOLE COMMAND' in line: mode = 'command'
    
    if mode is None: continue
    
    if line.startswith('['):
        name = line[1:-1].strip()        
        outDir = os.path.join(component, mode)
        path = os.path.join(outDir, name + '.ame')
        print path
        if not os.path.exists(outDir): os.makedirs(outDir)
        output = file(path, 'w')
        
    if output != None:
        if line.startswith('desc'):
            desc = line[line.index('=') + 1:].strip()
            print >> output, "@summary{\n    %s\n}" % escaped(desc)
        elif line.startswith('inf'):
            info = line[line.index('=') + 1:].strip()
            print >> output, "@description{\n    %s\n}" % escaped(info)
    


#-*- coding: UTF-8 -*-

"""
Some fancy helper functions.
"""

import os
import ctypes
from ctypes import POINTER
import operator
import numpy
from numpy import linalg

import logging;logger = logging.getLogger("pyassimp")

from .errors import AssimpError

additional_dirs, ext_whitelist = [],[]

# populate search directories and lists of allowed file extensions
# depending on the platform we're running on.
if os.name=='posix':
    additional_dirs.append('/usr/lib/')
    additional_dirs.append('/usr/local/lib/')

    # note - this won't catch libassimp.so.N.n, but 
    # currently there's always a symlink called
    # libassimp.so in /usr/local/lib.
    ext_whitelist.append('.so')
    # libassimp.dylib in /usr/local/lib
    ext_whitelist.append('.dylib')

elif os.name=='nt':
    ext_whitelist.append('.dll')
    path_dirs = os.environ['PATH'].split(';')
    for dir_candidate in path_dirs:
        if 'assimp' in dir_candidate.lower():
            additional_dirs.append(dir_candidate)
            
#print(additional_dirs)
def vec2tuple(x):
    """ Converts a VECTOR3D to a Tuple """
    return (x.x, x.y, x.z)

def transform(vector3, matrix4x4):
    """ Apply a transformation matrix on a 3D vector.

    :param vector3: a numpy array with 3 elements
    :param matrix4x4: a numpy 4x4 matrix
    """
    return numpy.dot(matrix4x4, numpy.append(vector3, 1.))

   
def get_bounding_box(scene):
    bb_min = [1e10, 1e10, 1e10] # x,y,z
    bb_max = [-1e10, -1e10, -1e10] # x,y,z
    return get_bounding_box_for_node(scene.rootnode, bb_min, bb_max, linalg.inv(scene.rootnode.transformation))

def get_bounding_box_for_node(node, bb_min, bb_max, transformation):

    transformation = numpy.dot(transformation, node.transformation)
    for mesh in node.meshes:
        for v in mesh.vertices:
            v = transform(v, transformation)
            bb_min[0] = min(bb_min[0], v[0])
            bb_min[1] = min(bb_min[1], v[1])
            bb_min[2] = min(bb_min[2], v[2])
            bb_max[0] = max(bb_max[0], v[0])
            bb_max[1] = max(bb_max[1], v[1])
            bb_max[2] = max(bb_max[2], v[2])


    for child in node.children:
        bb_min, bb_max = get_bounding_box_for_node(child, bb_min, bb_max, transformation)

    return bb_min, bb_max



def try_load_functions(library,dll,candidates):
    """try to  functbind to aiImportFile and aiReleaseImport
    
    library - path to current lib
    dll - ctypes handle to it
    candidates - receives matching candidates

    They serve as signal functions to detect assimp,
    also they're currently the only functions we need.
    insert (library,aiImportFile,aiReleaseImport,dll) 
    into 'candidates' if successful.

    """
    try:
        load = dll.aiImportFile
        release = dll.aiReleaseImport
    except AttributeError:
        #OK, this is a library, but it has not the functions we need
        pass
    else:
        #Library found!
        from .structs import Scene
        load.restype = POINTER(Scene)
        
        candidates.append((library, load, release, dll))


def search_library():
    """Loads the assimp-Library.
    
    result (load-function, release-function)    
    exception AssimpError if no library is found
    
        """
    #this path
    folder = os.path.dirname(__file__)

    # silence 'DLL not found' message boxes on win
    try:
        ctypes.windll.kernel32.SetErrorMode(0x8007)
    except AttributeError:
        pass    

    candidates = []
    
    # test every file
    for curfolder in [folder]+additional_dirs:
        for filename in os.listdir(curfolder):
            # our minimum requirement for candidates is that
            # they should contain 'assimp' somewhere in 
            # their name
            if filename.lower().find('assimp')==-1 or\
                os.path.splitext(filename)[-1].lower() not in ext_whitelist:
                continue

            library = os.path.join(curfolder, filename)
            logger.debug('Try ' + library)
            try:
                dll = ctypes.cdll.LoadLibrary(library)
            except Exception as e:
                logger.warning(str(e))
                # OK, this except is evil. But different OSs will throw different
                # errors. So just ignore any errors.
                continue

            try_load_functions(library,dll,candidates)
    
    if not candidates:
        # no library found
        raise AssimpError("assimp library not found")
    else:
        # get the newest library
        candidates = map(lambda x: (os.lstat(x[0])[-2], x), candidates)
        res = max(candidates, key=operator.itemgetter(0))[1]
        logger.debug('Using assimp library located at ' + res[0])

        # XXX: if there are 1000 dll/so files containing 'assimp'
        # in their name, do we have all of them in our address
        # space now until gc kicks in?

        # XXX: take version postfix of the .so on linux?
        return res[1:]

def hasattr_silent(object, name):
    """
        Calls hasttr() with the given parameters and preserves the legacy (pre-Python 3.2)
        functionality of silently catching exceptions.
        
        Returns the result of hasatter() or False if an exception was raised.
    """
    
    try:
        return hasattr(object, name)
    except:
        return False



# The arguments:
# 1 - output file (+".in" for input file)
# 2 - major
# 3 - minor
# 4 - patch
# 5 - label
# 6 - build number (optional)

import sys
import string
import time

def todays_build():
    now = time.localtime()
    return str((now.tm_year - 2011)*365 + now.tm_yday)

inFile = file(sys.argv[1] + '.in', 'rt')
outFile = file(sys.argv[1], 'wt')

if len(sys.argv) > 6:
    BUILD = sys.argv[6]
else:
    BUILD = todays_build()

VERSION = string.join(sys.argv[2:5], '.')
WIN_VERSION = string.join(sys.argv[2:5], ',') + ',' + BUILD

# Read in the template, substituting symbols for the correct info.
for line in inFile.readlines():
    def quoted(s):
        if line.strip() and line.strip()[0] == '"':
            return '""%s""' % s
        else:
            return '"%s"' % s
            
    line = line \
        .replace('LIBDENG2_VERSION_NUMBER', WIN_VERSION) \
        .replace('LIBDENG2_DESC_WSTR', quoted('Doomsday 2 core library')) \
        .replace('LIBDENG2_DESC', quoted('Doomsday 2 core library')) \
        .replace('LIBDENG2_VERSION_TEXT', quoted(VERSION)) \
        .replace('LIBDENG2_NICENAME', quoted('libcore')) \
        .replace('LIBDENG2_COPYRIGHT', quoted('2004-2014 Deng Team')) \
        .replace('LIBDENG2_FILENAME', quoted('deng_core.dll'))
        
    outFile.write(line)



#!/usr/bin/env python2.7
# coding=utf-8
#
# Repository Codex Generator by <jaakko.keranen@iki.fi>
# Generates a set of web pages out of tags in Git commit headlines.
# License: GNU GPL version 2 (or later)

import os, sys, time, string, base64

OUT_DIR = 'codex'
TITLE = 'Doomsday Codex'

if sys.argv > 1:
    OUT_DIR = sys.argv[1]

class Commit:
    def __init__(self, subject, author, date, link, hash):
        self.subject = subject
        self.author = author
        self.date = date
        self.link = link
        self.hash = hash
        self.extract_tags()
        
    def add_tag(self, tag):
        if len(tag) <= 1: return
        if tag.lower() not in map(lambda x: x.lower(), self.tags):
            self.tags.append(tag)
        
    def extract_tags(self):
        self.tags = []
        
        sub = self.subject
        
        if sub.startswith('Revert "') and sub[-1] == '"':
            sub = sub[8:-1]
            self.tags.append(u'Revert')
                        
        self.subject = sub[sub.find(': ') + 1:].strip()
                    
        # Some tags are automatically applied based on the subject line.
        sensitiveWords = self.subject.replace(';', ' ').replace(':', ' ').replace(',', ' ').split(' ')
        words = map(lambda w: w.lower(), sensitiveWords)
        if 'fixed' in words: self.add_tag(u'Fixed')
        if 'cleanup' in words: self.add_tag(u'Cleanup')
        if 'debug' in words: self.add_tag(u'Debug')
        if 'fixed' in words: self.add_tag(u'Fixed')
        if 'added' in words: self.add_tag(u'Added')
        if 'refactor' in self.subject.lower(): self.add_tag(u'Refactor')
        if 'GL' in sensitiveWords: self.add_tag(u'GL')
        if 'performance' in words: self.add_tag(u'Performance')
        if 'optimize' in words or 'optimized' in words or 'optimizing' in words: 
            self.add_tag(u'Optimize')

        # Are there any actual tags specified?
        if sub.find(': ') < 0: return
                                    
        sub = sub[:sub.find(': ')].replace(',', '|').replace('/', '|').replace('(', '|').\
            replace('&', '|').replace(')', '').replace('Changed ', 'Changed|')
                
        for unclean in sub.split('|'):
            unclean = unclean.replace('*nix', 'Unix')
            tag = unclean.replace('[', '').replace(']', '').replace('*', '').replace('"', '').strip()
            if 'ccmd' in tag.lower() or 'cvar' in tag.lower(): tag = u'Console'
            if 'cleanup' in tag.lower() or 'clean up' in tag.lower(): tag = u'Cleanup'
            if '_' in tag: continue
            if tag.lower() == 'chex':
                tag = u'Chex Quest'
            if tag.lower() == 'hacx':
                tag = u'HacX'
            if 'common' in tag.lower():
                tag = u'libcommon'
            if 'doom64' in tag.lower():
                tag = u'Doom64'
            elif 'doom' in tag.lower():
                tag = u'Doom'
            elif 'heretic' in tag.lower():
                tag = u'Heretic'
            elif 'hexen' in tag.lower():
                tag = u'Hexen'            
            if tag.lower() == 'add-on repository' or tag.lower() == 'addon repository':
                tag = u'Add-on Repository'
            if tag.lower() == 'bsp builder' or tag.lower() == 'bspbuilder':
                tag = u'BSP Builder'
            if tag.lower() == 'build repository' or tag.lower() == 'buildrepository':
                tag = u'Build Repository'
            if tag.lower() == 'deh reader' or tag.lower() == 'dehreader' or \
                    tag.lower() == 'deh read' or tag.lower() == 'dehacked reader' or \
                    tag.lower() == 'dehread':
                tag = u'Deh Reader'
            if tag.lower() == 'dpdehread':
                tag = u'dpDehRead'
            if tag.lower() == 'gcc':
                tag = u'GCC'
            if tag.lower() == 'glsandbox':
                tag = u'GLSandbox'
            if tag.lower() == 'dsdirectsound':
                tag = u'dsDirectSound'
            if tag.lower() == 'busy mode':
                tag = u'Busy Mode'
            if tag.lower() == 'dedicated server':
                tag = u'Dedicated Server'
            if tag.lower() == 'cmake':
                tag = u'CMake'
            if tag.lower().startswith('filesys') or tag.lower().startswith('file sys'):
                tag = u'File System'
            if tag.lower() == 'clang':
                tag = u'Clang'
            if 'zone' in tag.lower():
                tag = u'Memory Zone'
            if tag.lower() == 'infine':
                tag = u'InFine'
            if 'all games' in tag.lower():
                self.add_tag(u'All Games')
            if tag.lower() == 'lightgrid':
                tag = u'Light Grid'
            if tag.lower() == 'render':
                tag = u'Renderer'
            if tag.lower() == 'sfx':
                tag = u'SFX'
            if tag.lower() == 'taskbarwidth' or tag.lower() == 'taskbar' or tag.lower() == 'taskbarwidget':
                tag = u'Task Bar'
            if tag.lower() == 'texture manager':
                tag = u'Texture Manager'
            if 'refactoring' in tag.lower() or 'refactored' in tag.lower():
                tag = u'Refactor'
            if 'optimization' in tag.lower() or 'optimize' in tag.lower():
                tag = u'Optimize'
            if tag == 'osx' or tag.lower() == 'mac' or tag.lower() == 'mac os x':
                tag = u'OS X'
            if tag.lower().startswith('fixed'):
                tag = tag[5:].strip()
                self.add_tag(u'Fixed')
            if tag.lower().startswith('added'):
                tag = tag[5:].strip()
                self.add_tag(u'Added')
            if len(tag) > 2 and (tag[0] == '(' and tag[-1] == ')'):
                tag = tag[1:-1]
            if tag == u'WadMapConverter':
                tag = u'Wad Map Converter'
            if tag.lower() == 'win32':
                tag = u'Windows'
            self.add_tag(tag)        
        
def contains_tag(subject):
    pos = subject.find(': ')
    if pos < 0:
        # It might still have some recognized words.
        for allowed in ['debug', 'optimize', 'cleanup', 'fixed', 'added', 'refactor']:
            if allowed in subject.lower():
                return True
    if pos > 60: return False
    for badWord in ['related ', ' up ', ' the ', ' a ', ' in ', ' of', ' on ', 
                    ' and ', ' then ', ' when', ' to ',  ' for ', ' with ', ' into ']:
        if badWord in subject[:pos]:
            return False
    if subject[pos-4:pos+1] == 'TODO:': return False
    # It can only contain colons after the tag marker.
    p = subject.find(':')
    if p >= 0 and p < pos: return False
    p = subject.find('.')
    if p >= 0 and p < pos: return False
    return True

def fetch_commits():
    commits = {}

    tmpName = '__ctmp'
    format = '[[Subject]]%s[[/Subject]]' + \
             '[[Author]]%an[[/Author]]' + \
             '[[Date]]%ai[[/Date]]' + \
             '[[Link]]http://github.com/skyjake/Doomsday-Engine/commit/%H[[/Link]]' + \
             '[[Hash]]%H[[/Hash]]'
    os.system("git log --format=\"%s\" >> %s" % (format, tmpName))
    logText = unicode(file(tmpName, 'rt').read(), 'utf-8')
    os.remove(tmpName)        

    pos = 0
    while True:    
        pos = logText.find('[[Subject]]', pos)
        if pos < 0: break # No more.
        end = logText.find('[[/Subject]]', pos)        
        subject = logText[pos+11:end]

        # Author.
        pos = logText.find('[[Author]]', pos)
        end = logText.find('[[/Author]]', pos)
        author = logText[pos+10:end]
        
        # Date.
        pos = logText.find('[[Date]]', pos)
        end = logText.find('[[/Date]]', pos)
        date = logText[pos+8:end]
        
        # Link.
        pos = logText.find('[[Link]]', pos)
        end = logText.find('[[/Link]]', pos)
        link = logText[pos+8:end]

        # Hash.
        pos = logText.find('[[Hash]]', pos)
        end = logText.find('[[/Hash]]', pos)
        hash = logText[pos+8:end]
        
        if contains_tag(subject):        
            commits[hash] = Commit(subject, author, date, link, hash)

    return commits

# Compile the commit database.
byHash = fetch_commits()
byTag = {}
for commit in byHash.values():
    # Index all commits by tag.
    for tag in commit.tags:
        if tag in byTag:
            if commit not in byTag[tag]:
                byTag[tag].append(commit)
        else:
            byTag[tag] = [commit]
          
relatedTags = [
    ['Fixed', 'Added', 'Refactor', 'Optimize', 'Revert', 'Cleanup', 'Debug'],
    ['Windows', 'OS X', 'Linux', 'Unix', 'Debian', 'Ubuntu', 'FreeBSD', 'X11', '64-bit'],
    ['Windows', 'Windows *'],
    ['Qt', 'SDL'],
    ['Builder', 'Builds', 'qmake', 'Project*', 'CMake', 'Distrib', 'GCC', 'MSVC', 'Clang', 'Git', 
     'TextMate', 'wikidocs'],
    ['Test*', 'GLSandbox'],
    ['Docs', 'Documentation', 'Codex', 'Readme', 'Doxygen', 'Amethyst'],
    ['Homepage', 'Add-on Repository', 'Build Repository', 'CSS', 'RSS', 'DEW', 'Forums'],
    ['libdeng2', 'libcore', 'App', 'Config', 'Log', 'LogBuffer', 'Widgets', 'Garbage', 'Garbages'],
    ['libgui', 'DisplayMode', 'GL', 'OpenGL', 'GL*', 'Atlas*', 'Image', '*Bank', 'Font'],
    ['libdeng', 'libdeng1', 'liblegacy', 'Str'],
    ['libshell', 'Shell', 'AbstractLineEditor', 'Link'],    
    ['Client', 'Client UI', 'UI', 'Console', 'Control Panel', 'Default Style', 'GameSelectionWidget',
     'Task Bar', 'Updater', 'WindowSystem', 'Client*'],
    ['Server', 'Dedicated Server', 'Dedicated Mode', 'Server*'],
    ['Snowberry', 'Shell', 'Amethyst', 'md2tool'],
    ['All Games', 'Doom', 'Heretic', 'Hexen', 'Doom64', 'Chex Quest', 'HacX'],
    ['Plugins', 'Wad Map Converter', 'Deh Reader', 'dsDirectSound', 'OpenAL', 'dpDehRead',
     'dsWinMM', 'Dummy Audio', 'Example Plugin', 'exampleplugin', 'FluidSynth', 'FMOD'],
    ['API', 'DMU', 'DMU API'],
    ['DED', 'DED Parser', 'Ded Reader', 'Definitions', 'Info', 'ScriptedInfo'],
    ['Ring Zero', 'GameSelectionWidget'],
    ['Script*', 'scriptsys', 'Script', 'Record', 'Variable'],
    ['File System', 'Folder', 'Feed', 'FS', 'FS1', 'File*', '*File'],
    ['Resource*', 'Material*', 'Texture*', 'Uri'],
    ['Renderer', '* Renderer', 'Model*', 'Light Grid'],
    ['Network', 'Multiplayer', 'Server', 'Protocol'],
    ['Concurrency', 'Task', 'libdeng2'],
    ['libcommon', 'Game logic', 'Menu', 'Game Menu', 'Game Save', 'Games', 'Automap'],
    ['World', 'GameMap', 'Map', 'BSP Builder', 'HEdge', 'Bsp*', 'Line*', 'Sector', 'SectionEdge', 'Wall*',
     'Blockmap', 'Polyobj', 'Plane'],
    ['Finale Interpreter', 'Finales', 'InFine'],
    ['Input*', 'Bindings', 'Joystick', 'MouseEvent', 'KeyEvent', 'libgui'],
    ['Audio*', 'SFX', 'Music', 'FluidSynth'],
    ['Widget*', '*Widget', '*Rule', 'Rule*', 'Animation' ]
]

def find_related_tags(tag):
    rels = []
    for group in relatedTags:
        # Any wildcards?
        tags = []
        for t in group:
            if '*' not in t:
                tags.append(t)
            elif t[0] == '*':
                for x in byTag.keys():
                    if x.endswith(t[1:]):
                        tags.append(x)
            elif t[-1] == '*':
                for x in byTag.keys():
                    if x.startswith(t[:-1]):
                        tags.append(x)                
        if tag not in tags: continue
        for t in tags:
            if t != tag and t not in rels:
                rels.append(t)            
    return sorted(rels, key=lambda s: s.lower())
          
if not os.path.exists(OUT_DIR): os.mkdir(OUT_DIR)          

def tag_filename(tag):
    return base64.urlsafe_b64encode(tag)

def encoded_text(logText):
    logText = logText.replace(u'&', u'&amp;')
    logText = logText.replace(u'ä', u'&auml;')
    logText = logText.replace(u'ö', u'&ouml;')
    logText = logText.replace(u'Ä', u'&Auml;')
    logText = logText.replace(u'Ö', u'&Ouml;')
    logText = logText.replace(u'<', u'&lt;')
    logText = logText.replace(u'>', u'&gt;')
    logText = filter(lambda c: c in string.whitespace or c > ' ', logText)
    return logText
                                                
def print_header(out, pageTitle):
    print >> out, '<!DOCTYPE html>'
    print >> out, '<html>'
    print >> out, '<head>'
    print >> out, '<meta charset="UTF-8">'
    print >> out, '<title>%s - %s</title>' % (pageTitle, TITLE)
    print >> out, '<style type="text/css">'
    print >> out, 'body, table { font-family: "Helvetica Neue", sans-serif; }'
    print >> out, 'a { color: black; text-decoration: none; }'
    print >> out, 'a:hover { text-decoration: underline; }'
    print >> out, '.skiplist { line-height: 150%; }'
    print >> out, 'table { width: 100%; }'
    print >> out, 'table td { vertical-align: top; }'
    print >> out, 'td.commit-date { width: 100px; padding-left: 1.5ex; }'
    print >> out, 'td.commit-tags { width: 200px; text-align: right; }'
    print >> out, '.footer { text-align: center; font-size: 80%; }'
    print >> out, '.end-symbol { font-size: 125%; }'
    print >> out, '</style>'
    print >> out, '</head>'
    print >> out, '<body>'
    print >> out, '<form id="findform" name="find" action="find_tag.php" method="get"><a href="index.html">Alphabetical Index</a> | <a href="index_by_size.html">Tags by Size</a> | Find tag: <input type="text" name="tag"> <input type="submit" value="Find"></form>'
    print >> out, '<h1>%s</h1>' % pageTitle
    
def print_footer(out):
    print >> out, '<div class="footer"><p class="end-symbol">&oplus;</p><p>This is the <a href="http://dengine.net/">Doomsday Engine</a> <a href="http://github.com/skyjake/Doomsday-Engine/">source code repository</a> commit tag index.</p><p>Last updated: %s</p></div>' % time.asctime()
    print >> out, '</body>'
    print >> out, '</html>'

def print_table(out, cells, cell_printer, numCols=4, tdStyle='', separateByFirstLetter=False, letterFunc=None):
    colSize = (len(cells) + numCols - 1) / numCols
    tdElem = '<td style="width: %i%%; %s">' % (100/numCols, tdStyle)
    print >> out, '<table><tr>' + tdElem

    idx = 0
    inCol = 0
    letter = ''
    
    while idx < len(cells):
        if separateByFirstLetter:
            if inCol > 0 and letter != '' and letter != letterFunc(cells[idx]):
                print >> out, '<br/>'
            letter = letterFunc(cells[idx])
            
        cell_printer(out, cells[idx])
    
        idx += 1
        inCol += 1
        if inCol == colSize:
            print >> out, tdElem
            inCol = 0
    
    print >> out, '</table>'
            
# Create the index page with all tags sorted alphabetically.
out = file(os.path.join(OUT_DIR, 'index.html'), 'wt')
print_header(out, 'Alphabetical Tag Index')

sortedTags = sorted(byTag.keys(), cmp=lambda a, b: cmp(a.lower(), b.lower()))

def alpha_index_cell(out, tag):
    count = len(byTag[tag])
    if count > 100:
        style = 'font-weight: bold'
    elif count < 5:
        style = 'color: #aaa'
    else:
        style = ''
    print >> out, '<a href="tag_%s.html"><span style="%s">%s (%i)</span></a><br/>' % (
        tag_filename(tag), style, encoded_text(tag), count)

print_table(out, sortedTags, alpha_index_cell, separateByFirstLetter=True,
            letterFunc=lambda t: t[0].lower())

print_footer(out)

# Create the index page with tags sorted by size.
out = file(os.path.join(OUT_DIR, 'index_by_size.html'), 'wt')
print_header(out, 'All Tags by Size')

def tag_size_comp(a, b):
    c = cmp(len(byTag[b]), len(byTag[a]))
    if c == 0: return cmp(a.lower(), b.lower())
    return c

def size_index_cell(out, tag):
    count = len(byTag[tag])
    if count > 100:
        style = 'font-weight: bold'
    elif count < 5:
        style = 'color: #aaa'
    else:
        style = ''
    print >> out, '<a href="tag_group_%s.html"><span style="%s">%i %s</span></a><br/>' % (
        tag_filename(tag), style, count, encoded_text(tag))    

print_table(out, sorted(byTag.keys(), cmp=tag_size_comp), size_index_cell)

print_footer(out)

def print_tags(out, com, tag, linkSuffix=''):
    first = True
    for other in sorted(com.tags, cmp=lambda a, b: cmp(a.lower(), b.lower())):
        if other != tag:
            if not first: print >> out, '| '
            first = False
            print >> out, '<b><a href="tag_%s%s.html">%s</a></b>' % (linkSuffix, \
                    tag_filename(other), encoded_text(other))
    
def print_commit(out, com, tag, linkSuffix=''):
    print >> out, trElem + '<td class="commit-date"><a href="%s">' % com.link + com.date[:10] + '</a> '
    print >> out, '<td class="commit-tags">'
    print_tags(out, com, tag, linkSuffix)
    print >> out, '<td>:'
    print >> out, ('<td><a href="%s">' % com.link + encoded_text(com.subject)[:250] + '</a>').encode('utf8')

colors = ['#f00', '#0a0', '#00f', 
          '#ed0', '#f80', '#0ce', 
          '#88f', '#b0b',  
          '#f99', '#8d8', '#222',
          '#44a', '#940', '#888']

def html_color(colIdx):
    if colIdx == None: return '#000'
    return colors[colIdx % len(colors)]

def color_box(colIdx):
    return ' <span style="padding-left:1em; background-color:%s">&nbsp;</span> ' % \
        html_color(colIdx)

def print_date_sorted_commits(out, coms, tag, linkSuffix='', colorIdx=None):
    months = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August',
              'September', 'October', 'November', 'December']
    curDate = ''
    dateSorted = sorted(coms, key=lambda c: c.date, reverse=True)
    print >> out, '<table>'
    for com in dateSorted:
        # Monthly headers.
        if curDate == '' or curDate != com.date[:7]:
            print >> out, '<tr><td colspan="4" style="padding-left:1ex; padding-top:1.5em; padding-bottom:0.5em; font-weight:bold; font-size:110%%; color:%s; border-bottom:1px dotted %s">' % \
                (html_color(colorIdx), html_color(colorIdx))
            print >> out, months[int(com.date[5:7]) - 1], com.date[:4]
        curDate = com.date[:7]        
        print_commit(out, com, tag, linkSuffix)
    print >> out, '</table>'

# Create the tag redirecting PHP page.
out = file(os.path.join(OUT_DIR, 'find_tag.php'), 'wt')
print >> out, '<?php'

print >> out, '$tags = array(', string.join(['"%s" => "%s"' % (tag, tag_filename(tag)) for tag in byTag.keys()], ', '), ');'
print >> out, """
$input = stripslashes(strip_tags($_GET["tag"]));
$style = $_GET["style"];
if($style == 'grouped') {
  $style = 'group_';
} else {
  $style = '';
}
$destination = "index.html";
$best = -1.0;
if(strlen($input) > 0 && strlen($input) < 60) {
  foreach($tags as $tag => $link) {
    $lev = (float) levenshtein($input, $tag); // case sensitive
    if(stripos($tag, $input) !== FALSE || stripos($input, $tag) !== FALSE) {
        // Found as a case insensitive substring, increase likelihood.
        $lev = $lev/2.0;
    }
    if(stripos($tag, $input) === 0) {
        // Increase likelihood further if the match is in the beginning.
        $lev = $lev/2.0;
    }
    if(!strcasecmp($tag, $input) == 0) {
        // Case insensitive direct match, increase likelihood.
        $lev = $lev/2.0;
    }
    if($lev == 0) {
      $destination = "tag_$style$link.html";
      break;
    }
    if($best < 0 || $lev < $best) {
      $destination = "tag_$style$link.html";
      $best = $lev;
    }
  }
}
header("Location: $destination");
"""

print >> out, "?>"

def print_related_tags(out, tag, style=''):
    rels = find_related_tags(tag)
    if len(rels) == 0: return
    print >> out, '<p>Related tags: '
    print >> out, string.join(['<a href="tag_%s%s.html">%s</a>' % (style, tag_filename(t), t)
                               for t in rels], ', ')
    print >> out, '</p>'

def percentage(part, total):
    return int(round(float(part)/float(total) * 100))

def print_authorship(out, tag):
    authors = {}
    total = len(byTag[tag])
    for commit in byTag[tag]:    
        if commit.author in authors:
            authors[commit.author] += 1
        else:
            authors[commit.author] = 1
    sortedAuthors = sorted(authors.keys(), key=lambda a: authors[a], reverse=True)    
    print >> out, '<p>Authorship:', string.join(['%i%% %s' % (percentage(authors[a], total), a) 
                                                 for a in sortedAuthors], ', ').encode('utf8'), '</p>'
    
#
# Create pages for each tag.
#
for tag in byTag.keys():
    #print 'Generating tag page:', tag

    trElem = '<tr style="padding:1ex">'
    
    # First a simple date-based list of commits in this tag.
    out = file(os.path.join(OUT_DIR, 'tag_%s.html' % tag_filename(tag)), 'wt')
    print_header(out, tag)
    print_related_tags(out, tag)
    print_authorship(out, tag)
    print >> out, '<p><a href="tag_group_%s.html"><b>View commits by groups</b></a></p>' % tag_filename(tag)
    print >> out, '<div style="margin-left:1em">'
    print_date_sorted_commits(out, byTag[tag], tag)
    print >> out, '</div>'
    print_footer(out)
    
    present = {}
    for com in byTag[tag]:
        for other in com.tags:
            if other == tag: continue
            if other not in present:
                present[other] = [com]
            else:
                present[other].append(com)

    def present_by_size(a, b):
        c = cmp(len(present[b]), len(present[a]))
        if c == 0: return cmp(a.lower(), b.lower())
        return c
        
    presentSorted = sorted(present.keys(), cmp=present_by_size)
    
    # Then grouped by subgroup size.            
    out = file(os.path.join(OUT_DIR, 'tag_group_%s.html' % tag_filename(tag)), 'wt')
    print_header(out, tag + ' (Grouped)')
    print_related_tags(out, tag, 'group_')
    print_authorship(out, tag)
    print >> out, '<p><a id="top"></a><a href="tag_%s.html"><b>View commits by date</b></a></p>' % tag_filename(tag)

    if len(byTag[tag]) > 10:
        print >> out, '<div class="skiplist"><b>Jump down to:</b>'

        class SkipTagPrinter:
            def __init__(self):
                self.color = 0
            
            def __call__(self, out, tag):
                print >> out, '<span style="white-space:nowrap;">' + color_box(self.color) 
                self.color += 1
                print >> out, '<a href="#%s">%s (%i)</a></span><br/>' % (tag_filename(tag),
                    tag, len(present[tag]))
    
        print_table(out, presentSorted, SkipTagPrinter(), numCols=5, tdStyle='line-height:150%')

        print >> out, '</div>'
    
    # First the commits without any subgroups.
    print >> out, '<div style="margin-left:1em">'
    print_date_sorted_commits(out, filter(lambda c: len(c.tags) == 1, byTag[tag]), tag)    
    print >> out, '</div>'

    color = 0
    for tag2 in presentSorted:
        print >> out, '<h2><a id="%s"></a>%s<a href="tag_group_%s.html">%s</a> (%i) <span style="color:#ccc">&mdash; %s</span></h2>' % (
            tag_filename(tag2), color_box(color), tag_filename(tag2), tag2, len(present[tag2]), tag)
        print >> out, '<div style="margin-left:0.85em; border-left:4px solid %s">' % html_color(color)
        print_date_sorted_commits(out, filter(lambda c: tag2 in c.tags, byTag[tag]), tag, 'group_',
            colorIdx=color)
        print >> out, '</div>'
        print >> out, '<p><small><a href="#top">&uarr; Back to top</a></small></p>'
        color += 1
    
    print_footer(out)



#!/usr/bin/env python
#-*- coding: UTF-8 -*-

""" This program demonstrates the use of pyassimp to load and
render objects with OpenGL.

'c' cycles between cameras (if any available)
'q' to quit

This example mixes 'old' OpenGL fixed-function pipeline with 
Vertex Buffer Objects.

Materials are supported but textures are currently ignored.

For a more advanced example (with shaders + keyboard/mouse 
controls), check scripts/sdl_viewer.py

Author: Séverin Lemaignan, 2012

This sample is based on several sources, including:
 - http://www.lighthouse3d.com/tutorials
 - http://www.songho.ca/opengl/gl_transform.html
 - http://code.activestate.com/recipes/325391/
 - ASSIMP's C++ SimpleOpenGL viewer
"""

import os, sys
from OpenGL.GLUT import *
from OpenGL.GLU import *
from OpenGL.GL import *

import logging;logger = logging.getLogger("pyassimp_opengl")
logging.basicConfig(level=logging.INFO)

import math
import numpy

import pyassimp
from pyassimp.postprocess import *
from pyassimp.helper import *


name = 'pyassimp OpenGL viewer'
height = 600
width = 900

class GLRenderer():
    def __init__(self):

        self.scene = None

        self.using_fixed_cam = False
        self.current_cam_index = 0

        # store the global scene rotation
        self.angle = 0.

        # for FPS calculation
        self.prev_time = 0
        self.prev_fps_time = 0
        self.frames = 0

    def prepare_gl_buffers(self, mesh):
        """ Creates 3 buffer objets for each mesh, 
        to store the vertices, the normals, and the faces
        indices.
        """

        mesh.gl = {}

        # Fill the buffer for vertex positions
        mesh.gl["vertices"] = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, mesh.gl["vertices"])
        glBufferData(GL_ARRAY_BUFFER, 
                    mesh.vertices,
                    GL_STATIC_DRAW)

        # Fill the buffer for normals
        mesh.gl["normals"] = glGenBuffers(1)
        glBindBuffer(GL_ARRAY_BUFFER, mesh.gl["normals"])
        glBufferData(GL_ARRAY_BUFFER, 
                    mesh.normals,
                    GL_STATIC_DRAW)


        # Fill the buffer for vertex positions
        mesh.gl["triangles"] = glGenBuffers(1)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh.gl["triangles"])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
                    mesh.faces,
                    GL_STATIC_DRAW)

        # Unbind buffers
        glBindBuffer(GL_ARRAY_BUFFER,0)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0)

    def load_model(self, path, postprocess = None):
        logger.info("Loading model:" + path + "...")

        if postprocess:
            self.scene = pyassimp.load(path, postprocess)
        else:
            self.scene = pyassimp.load(path)
        logger.info("Done.")

        scene = self.scene
        #log some statistics
        logger.info("  meshes: %d" % len(scene.meshes))
        logger.info("  total faces: %d" % sum([len(mesh.faces) for mesh in scene.meshes]))
        logger.info("  materials: %d" % len(scene.materials))
        self.bb_min, self.bb_max = get_bounding_box(self.scene)
        logger.info("  bounding box:" + str(self.bb_min) + " - " + str(self.bb_max))

        self.scene_center = [(a + b) / 2. for a, b in zip(self.bb_min, self.bb_max)]

        for index, mesh in enumerate(scene.meshes):
            self.prepare_gl_buffers(mesh)

        # Finally release the model
        pyassimp.release(scene)

    def cycle_cameras(self):
        self.current_cam_index
        if not self.scene.cameras:
            return None
        self.current_cam_index = (self.current_cam_index + 1) % len(self.scene.cameras)
        cam = self.scene.cameras[self.current_cam_index]
        logger.info("Switched to camera " + str(cam))
        return cam

    def set_default_camera(self):

        if not self.using_fixed_cam:
            glLoadIdentity()

            gluLookAt(0.,0.,3.,
                      0.,0.,-5.,
                      0.,1.,0.)



    def set_camera(self, camera):

        if not camera:
            return

        self.using_fixed_cam = True

        znear = camera.clipplanenear
        zfar = camera.clipplanefar
        aspect = camera.aspect
        fov = camera.horizontalfov

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        # Compute gl frustrum
        tangent = math.tan(fov/2.)
        h = znear * tangent
        w = h * aspect

        # params: left, right, bottom, top, near, far
        glFrustum(-w, w, -h, h, znear, zfar)
        # equivalent to:
        #gluPerspective(fov * 180/math.pi, aspect, znear, zfar)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        cam = transform(camera.position, camera.transformation)
        at = transform(camera.lookat, camera.transformation)
        gluLookAt(cam[0], cam[2], -cam[1],
                   at[0],  at[2],  -at[1],
                       0,      1,       0)

    def fit_scene(self, restore = False):
        """ Compute a scale factor and a translation to fit and center 
        the whole geometry on the screen.
        """

        x_max = self.bb_max[0] - self.bb_min[0]
        y_max = self.bb_max[1] - self.bb_min[1]
        tmp = max(x_max, y_max)
        z_max = self.bb_max[2] - self.bb_min[2]
        tmp = max(z_max, tmp)

        if not restore:
            tmp = 1. / tmp

        logger.info("Scaling the scene by %.03f" % tmp)
        glScalef(tmp, tmp, tmp)

        # center the model
        direction = -1 if not restore else 1
        glTranslatef( direction * self.scene_center[0], 
                      direction * self.scene_center[1], 
                      direction * self.scene_center[2] )

        return x_max, y_max, z_max

    def apply_material(self, mat):
        """ Apply an OpenGL, using one OpenGL display list per material to cache 
        the operation.
        """

        if not hasattr(mat, "gl_mat"): # evaluate once the mat properties, and cache the values in a glDisplayList.
            diffuse = numpy.array(mat.properties.get("diffuse", [0.8, 0.8, 0.8, 1.0]))
            specular = numpy.array(mat.properties.get("specular", [0., 0., 0., 1.0]))
            ambient = numpy.array(mat.properties.get("ambient", [0.2, 0.2, 0.2, 1.0]))
            emissive = numpy.array(mat.properties.get("emissive", [0., 0., 0., 1.0]))
            shininess = min(mat.properties.get("shininess", 1.0), 128)
            wireframe = mat.properties.get("wireframe", 0)
            twosided = mat.properties.get("twosided", 1)

            setattr(mat, "gl_mat", glGenLists(1))
            glNewList(mat.gl_mat, GL_COMPILE)
    
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse)
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular)
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient)
            glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, emissive)
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess)
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE if wireframe else GL_FILL)
            glDisable(GL_CULL_FACE) if twosided else glEnable(GL_CULL_FACE)
    
            glEndList()
    
        glCallList(mat.gl_mat)

    
   
    def do_motion(self):

        gl_time = glutGet(GLUT_ELAPSED_TIME)

        self.angle = (gl_time - self.prev_time) * 0.1

        self.prev_time = gl_time

        # Compute FPS
        self.frames += 1
        if gl_time - self.prev_fps_time >= 1000:
            current_fps = self.frames * 1000 / (gl_time - self.prev_fps_time)
            logger.info('%.0f fps' % current_fps)
            self.frames = 0
            self.prev_fps_time = gl_time

        glutPostRedisplay()

    def recursive_render(self, node):
        """ Main recursive rendering method.
        """

        # save model matrix and apply node transformation
        glPushMatrix()
        m = node.transformation.transpose() # OpenGL row major
        glMultMatrixf(m)

        for mesh in node.meshes:
            self.apply_material(mesh.material)

            glBindBuffer(GL_ARRAY_BUFFER, mesh.gl["vertices"])
            glEnableClientState(GL_VERTEX_ARRAY)
            glVertexPointer(3, GL_FLOAT, 0, None)

            glBindBuffer(GL_ARRAY_BUFFER, mesh.gl["normals"])
            glEnableClientState(GL_NORMAL_ARRAY)
            glNormalPointer(GL_FLOAT, 0, None)

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh.gl["triangles"])
            glDrawElements(GL_TRIANGLES,len(mesh.faces) * 3, GL_UNSIGNED_INT, None)

            glDisableClientState(GL_VERTEX_ARRAY)
            glDisableClientState(GL_NORMAL_ARRAY)

            glBindBuffer(GL_ARRAY_BUFFER, 0)
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)

        for child in node.children:
            self.recursive_render(child)

        glPopMatrix()


    def display(self):
        """ GLUT callback to redraw OpenGL surface
        """
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)

        glRotatef(self.angle,0.,1.,0.)
        self.recursive_render(self.scene.rootnode)

        glutSwapBuffers()
        self.do_motion()
        return

    ####################################################################
    ##               GLUT keyboard and mouse callbacks                ##
    ####################################################################
    def onkeypress(self, key, x, y):
        if key == 'c':
            self.fit_scene(restore = True)
            self.set_camera(self.cycle_cameras())
        if key == 'q':
            sys.exit(0)

    def render(self, filename=None, fullscreen = False, autofit = True, postprocess = None):
        """

        :param autofit: if true, scale the scene to fit the whole geometry
        in the viewport.
        """
    
        # First initialize the openGL context
        glutInit(sys.argv)
        glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
        if not fullscreen:
            glutInitWindowSize(width, height)
            glutCreateWindow(name)
        else:
            glutGameModeString("1024x768")
            if glutGameModeGet(GLUT_GAME_MODE_POSSIBLE):
                glutEnterGameMode()
            else:
                print("Fullscreen mode not available!")
                sys.exit(1)

        self.load_model(filename, postprocess = postprocess)


        glClearColor(0.1,0.1,0.1,1.)
        #glShadeModel(GL_SMOOTH)

        glEnable(GL_LIGHTING)

        glEnable(GL_CULL_FACE)
        glEnable(GL_DEPTH_TEST)

        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE)
        glEnable(GL_NORMALIZE)
        glEnable(GL_LIGHT0)

        glutDisplayFunc(self.display)


        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(35.0, width/float(height) , 0.10, 100.0)
        glMatrixMode(GL_MODELVIEW)
        self.set_default_camera()

        if autofit:
            # scale the whole asset to fit into our view frustum·
            self.fit_scene()

        glPushMatrix()

        glutKeyboardFunc(self.onkeypress)
        glutIgnoreKeyRepeat(1)

        glutMainLoop()


if __name__ == '__main__':
    if not len(sys.argv) > 1:
        print("Usage: " + __file__ + " <model>")
        sys.exit(0)

    glrender = GLRenderer()
    glrender.render(sys.argv[1], fullscreen = False, postprocess = aiProcessPreset_TargetRealtime_MaxQuality)




# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#   Antti Kopponen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file profdb.py Profile Management
##
## @todo The Defaults profile should have its own class.  A
## DefaultsProfile class could then cleanly handle such special cases
## as changing of the user interface language.

import os, re, string, shutil
import events, paths, cfparser, language, logger
import sb.profile
import ui


# The list of available profiles.  All of these will be displayed in the
# Profile List Box.
profiles = []

# The currently active profile.
active = None

# The Defaults profile.
defaults = None


def get(identifier):
    """Returns the profile with the specified identifier.

    @param identifier The identifier of the profile.

    @return A Profile object, or None if the identifier isn't in use.
    """
    for p in profiles:
        if p.getId() == identifier:
            return p
    return None


def getProfiles(func=None):
    return [p for p in profiles if not func or func(p)]


def refresh(hasChanged=False):
    """Send a notification that will cause everybody to refresh their
    state with regards to the active profile.
    
    @todo  Move this out of the module so the import is unnecessary?
    """
    ui.freeze()
    
    event = events.ActiveProfileNotify(active, hasChanged)
    events.send(event)

    ui.unfreeze() 


def setActive(profile):
    """Set a new active profile.  If the specified profile is not the
    same as the currently active profile, a notification is sent.

    @param profile A Profile object.
    """
    global active

    if profile is not active:
        active = profile
        refresh(True)


def getActive():
    return active


def getDefaults():
    return defaults


def save():
    """Write all the profiles to disk.  Also saves which of the profiles
    is currently active.
    """
    path = paths.getUserPath(paths.PROFILES)

    # First we'll remove all the ".prof"-files in the user's profile directory.
    #for name in os.listdir(path):
    #    fileName = os.path.join(path, name)
    #    if paths.hasExtension('prof', fileName):
    #        os.remove(fileName)

    # Then we'll create files for Profiles and fill them with information.
    for p in profiles:
        fileName = os.path.join(path, p.getFileName())

        f = file(fileName, 'w')

        # Name has to be written even in user's ".prof"-files.
        f.write('name: ' + p.getName() + "\n")

        # Hidden profiles don't show up in the profiles list.
        if p.isHidden():
            f.write("hidden: yes\n")

        # If the Profile happens to be Defaults, we'll write there
        # information concerning the active Profile.
        if p is defaults:
            f.write('active: ' + active.getId() + "\n")

        # The banner image.
        f.write('banner: ' + p.getBanner() + "\n")

        # The list of attached Addons is saved.
        addons = p.getAddons()
        if len(addons) > 0:
            f.write("addons <" + string.join(p.getAddons(), ', ') + ">\n")

        # Load order is saved.
        if len(p.getLoadOrder()):
            f.write("load-order <" + string.join(p.getLoadOrder(), ', ') +
                    ">\n")                    

        # Values are saved.
        for value in p.getAllValues():
            f.write(value.getId() + ': ' + value.getValue() + "\n")

        f.close()


def _listProfilesIn(locations):
    """Compose a list of profile file names by reading from the specified 
    locations.

    @param paths A list of directory paths.

    @return An array of absolute file names.
    """
    names = []

    for path in locations:
        try:
            for name in os.listdir(path):
                if paths.hasExtension('prof', name):
                    names.append(os.path.join(path, name))
        except OSError:
            # Such a location does not exist, ignore it.
            pass

    return names


def restore():
    """Reads all the profiles from disk.  Restores the currently active
    profile as well.  This function has to be called after starting
    the program before any profiles are accessed.

    System profiles that aren't present in the user's profile
    directory are copied to the user's profile directory.
    """
    global defaults
    global profiles
    global restoredActiveId

    # By default, restore selection to the Defaults profile.
    restoredActiveId = 'defaults'

    # Clear the current profile list.
    profiles = []

    systemProfilePaths = [paths.getSystemPath(paths.PROFILES)] + \
                          paths.getBundlePaths(paths.PROFILES)
    userProfilePath = paths.getUserPath(paths.PROFILES)

    # List all the system and user profiles.
    availSystem = _listProfilesIn(systemProfilePaths)
    availUser = _listProfilesIn([userProfilePath])

    # We are going to load only the profiles in the user's direcory,
    # but before that make sure that the user has an instance of each
    # system profile.
    for sysFile in availSystem:
        identifier = paths.getBase(sysFile)

        # Does this exist in the user's directory?
        gotIt = False
        for userFile in availUser:
            if paths.getBase(userFile) == identifier:
                gotIt = True
                break

        if not gotIt:
            # Since the system profile does not exist on the user,
            # copy it to the user profile directory.
            shutil.copyfile(sysFile,
                            os.path.join(userProfilePath,
                                         os.path.basename(sysFile)))

    # Find every profile in system's and user's profile directories.
    # Files in the user's directory augment the files in the system
    # directory.
    for name in _listProfilesIn([userProfilePath]):
        load(os.path.join(userProfilePath, name), False)
    logger.show()

    defaults = get('defaults')

    if defaults is None:
        # Recreate the Defaults profile.
        load(os.path.join(systemProfilePath, "defaults.prof"), False)
        defaults = get('defaults')
        logger.show()

    # Set the default language.
    lang = defaults.getValue('language')
    if lang:
        language.change(lang.getValue())

    # Send profile-loaded notifications.
    for p in profiles:
        if not p.isHidden():
            events.send(events.ProfileNotify(p))

    # Restore the previously active profile.
    prof = get(restoredActiveId)
    if prof:
        setActive(prof)
    else:
        setActive(defaults)


def load(path, notify=True):
    """Loads a single profile from disk.  After the profile has been
    loaded, sends a "profile-updated" notification so that the plugins
    may react to the new profile.  If a profile with the

    @param path The file name of the profile to load.
    """

    # Parameter "path" stands for profile's file name. Every time a
    # profile is being loaded, we'll have to check if the profile
    # allready exists. In that case, old profile will be updated by
    # new profile's information. Finally when the loading is complete,

    # The identifier of the profile (the base name of the file).
    identifier = paths.getBase(path)

    # Parse the file.
    elements = []
    p = cfparser.FileParser(path)
    try:
        while True:
            elements.append(p.get())

    except cfparser.OutOfElements:
        # The file ended normally.
        pass

    except cfparser.ParseFailed, ex:
        # Show the error dialog.
        logger.add(logger.HIGH, 'error-parsing-profile', path, str(ex))
        return
                
    except Exception:
        # All other errors also cause the error dialog to appear.
        logger.add(logger.HIGH, 'error-read-profile', path,
                   logger.formatTraceback())
        return

    # Should be use an existing profile?
    profile = get(identifier)

    if not profile:
        # Create a new blank profile.
        profile = sb.profile.Profile(identifier)
        # Add it to the main list of profiles.
        global profiles
        profiles.append(profile)

    # Process all the elements.
    for e in elements:
        if e.isKey() and e.getName() == 'name':
            # The user-provided name.
            profile.setName(e.getValue())
            language.define('english', identifier, profile.getName())

        elif e.isKey() and e.getName() == 'active':
            # The identifier of the active profile.  This is stored in
            # the Defaults profile and restore() uses it to return to
            # the active profile that was used when save() was called.
            global restoredActiveId
            restoredActiveId = e.getValue()

        elif e.isKey() and e.getName() == 'hidden':
            profile.setHidden(e.getValue() == 'yes')

        elif e.isList() and e.getName() == 'load-order':
            profile.setLoadOrder(e.getContents())

        elif e.isKey() and e.getName() == 'story':
            # A story associated with the profile.  An HTML-formatted
            # description of the profile, really.
            profile.setStory(e.getValue())

        elif e.isKey() and e.getName() == 'banner':
            profile.setBanner(e.getValue())

        elif e.isList() and e.getName() == 'addons':
            # The addons attached to the profile.
            profile.setAddons(e.getContents())

        elif e.isKey():
            # A value for a setting?
            try:
                # No notifications will be sent.
                profile.setValue(e.getName(), e.getValue(), False)
            except KeyError:
                # Setting does not exist, ignore it.
                pass

    if notify:
        # The profile has been loaded.
        events.send(events.ProfileNotify(profile))


def _getNewProfileId():
    # Choose the new identifier by comparing against the profiles
    # already loaded in memory.
    for i in range(10000):
        identifier = 'userprofile%i' % i
        if not get(identifier):
            # This is unused.
            break
    return identifier


def create(name, gameComponent):
    """Create a new profile with the specified name.  The identifier
    of the profile is generated automatically.  The identifier is not
    visible to the user.

    @param name The visible name of the new profile.

    @param gameComponent The game that the profile will be using.

    @return The Profile object that was created.
    """

    # Generate an unused identifier.
    identifier = _getNewProfileId()

    p = sb.profile.Profile(identifier)

    # Set the basic information of the profile.
    p.setName(name)
    p.setValue('game', gameComponent)

    # Add it to the list of profiles and send a notification.
    profiles.append(p)
    events.send(events.ProfileNotify(p))
    setActive(p)
    return p


def duplicate(identifier, name):
    if identifier == 'defaults':
        return

    # Generate an unused identifier.
    p = get(identifier).duplicate(_getNewProfileId())

    p.setName(name)

    # Add it to the list of profiles and send a notification.
    profiles.append(p)
    events.send(events.ProfileNotify(p))
    setActive(p)
    return p


def hide(identifier):
    """Mark a profile hidden."""

    prof = get(identifier)

    if prof is defaults:
        # Hiding the defaults is not possible.
        return

    prof.setHidden(True)
    events.send(events.ProfileNotify(prof))


def show(identifier):
    """Clear the hide flag and show a profile."""

    prof = get(identifier)
    prof.setHidden(False)
    events.send(events.ProfileNotify(prof))


def remove(identifier):
    """Removes one profile from profiles.

    @param identifier Indentifier of the profile to remove.
    """
    if identifier == 'defaults':
        # The Defaults profile can't be removed.
        return

    prof = get(identifier)

    hide(identifier)

    # System profiles can't be deleted, only hidden.
    if not prof.isSystemProfile():
        fileName = os.path.join(paths.getUserPath(paths.PROFILES),
                               prof.getFileName())
        if os.path.exists(fileName):
            os.remove(fileName)
        profiles.remove(prof)


def reset(identifier, resetValues=True, resetAddons=True):
    """Resetting a profile removes all values except the 'iwad' value.

    @param identifier Profile identifier.
    @param resetValues  True, if the profile's values should be reset.
    @param resetAddons  True, if the profile's addons should be reset.
    """
    prof = get(identifier)

    if resetValues:
        valueIds = map(lambda v: v.getId(), prof.getAllValues())
        for id in valueIds:
            # Some values cannot be reset.
            if id != 'iwad' and id != 'game' and id != 'setup-wizard-shown':
                prof.removeValue(id)

    if resetAddons:
        # Also detach all addons, except the Deathkings IWAD.
        for addon in prof.getAddons():
            # Huge kludge!  Reset must not render Deathkings unlaunchable.
            # There should be a better way to declare which addons are not
            # resetable.
            if identifier == 'hexen-dk' and addon == 'hexdd-wad':
                continue
        
            prof.removeAddon(addon)



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file paths.py File Path Services
##
## This module contains functions that can be used to determine where
## in the filesystem the various parts of the Snowberry application
## and data are located.
##
## It is assumed that the current working directory is Snowberry's
## installation directory, and also contains Doomsday.

import os, re, sys
import host, events, logger


def isHomeless():
    """Homeless setup means that Snowberry uses the main installation
    directory for user data.  In Windows, the default is homeless
    mode."""

    noHome = '-nohome' in sys.argv
    useHome = '-home' in sys.argv

    if useHome:
        return False

    if noHome:
        return True

    # The default is homeless in Windows versions earlier than Vista and
    # Win Server 2008.
    if host.isWindows() and not host.isWindowsVista():
        return True

    return False


# Directory Type constants to be used with getSystemPath() and
# getUserPath().
ADDONS = 'addons'
MANIFESTS = 'manifests'
UNINSTALLED = 'uninstalled'
CONF = 'conf'
LANG = 'lang'
PLUGINS = 'plugins'
PROFILES = 'profiles'
GRAPHICS = 'graphics'
RUNTIME = 'runtime'
SERVER_RUNTIME = 'server-runtime'

# List of custom addons directories.
addonPaths = []


# On the Mac, change the current directory to Resources inside the bundle.
if host.isMac():
    import __main__
    os.chdir(os.path.dirname(os.path.abspath(__main__.__file__)))

# This is the name of the directory under which Snowberry's files will
# be stored in the user's home directory.
if host.isMac():
    SNOWBERRY_HOME_DIR = 'Doomsday Engine'
elif host.isWindowsVista():
    SNOWBERRY_HOME_DIR = os.path.join('Documents', 'Doomsday Frontend')
elif host.isWindows():
    SNOWBERRY_HOME_DIR = os.path.join('My Documents', 'Doomsday Frontend')
else:
    SNOWBERRY_HOME_DIR = '.doomsday'
    
# In a homeless setup all the user data goes in the main installation folder.
if isHomeless():
    SNOWBERRY_HOME_DIR = '.'

# This is the full path of the user's Snowberry home directory.
homeDir = None

# Bundle paths.  Files will be searched in these after the user path,
# but before the system path.
bundlePaths = []

# The custom paths override the default user paths.
customPaths = {}


def _createDir(path):
    """Creates the specified directory if it doesn't already exist."""
    try:
        os.makedirs(path)
    except:
        # Must exist already.
        pass


def _checkSnowberryHome():
    """Checks if the Snowberry home directory exists.  If it doesn't, it
    is created and populated with the basic files and directories."""

    global homeDir

    # When this is run for the first time, we'll need to figure out
    # where the home directory should be.
    if not homeDir:
        if isHomeless():
            homeLocation = os.getcwd()           
        else:
            if host.isMac():
                # Home on the Mac.
                homeLocation = os.path.join(os.getenv('HOME'),
                                            'Library/Application Support')
            else:
                # First see if a HOME environment variable has been defined.
                homeLocation = os.getenv('HOME')

            if not homeLocation:
                # The failsafe.
                homeLocation = os.getcwd()
                # The environment variable was not defined.  Let's try
                # something else.
                if host.isWindows():
                    if os.getenv('HOMEPATH'):
                        homeLocation = os.getenv('HOMEDRIVE') + \
                                       os.getenv('HOMEPATH')
                    elif os.getenv('USERPROFILE'):
                        homeLocation = os.getenv('USERPROFILE')
                    elif os.getenv('APPDATA'):
                        homeLocation = os.getenv('APPDATA')

        homeDir = os.path.join(homeLocation, SNOWBERRY_HOME_DIR)
        
    homeDir = os.path.abspath(homeDir)

    # The home needs to be created if it doesn't exist yet.
    _createDir(homeDir)

    # Create the rest of the user directories if they don't exist yet.
    for p in [ADDONS, MANIFESTS, UNINSTALLED, CONF, LANG, PLUGINS, PROFILES,
              GRAPHICS, RUNTIME, SERVER_RUNTIME]:
        _createDir(getUserPath(p))


def isCaseSensitive():
    """Returns True if the file system is case sensitive."""

    return not host.isWindows() and not host.isMac()


def addBundlePath(path):
    """Add a path to the list of paths where files will be searched
    from.

    @param path The bundle contents path.
    """
    global bundlePaths
    bundlePaths.append(path)


def setCustomPath(path, custom):
    """Set a custom path.
    
    @param path The path to set (PROFILES, RUNTIME, CONF, etc.).
    
    @param custom The custom path to use.
    """
    customPaths[path] = custom


def getHomeDirectory():
    "Returns the current user's Snowberry home directory."
    return homeDir


def getSystemPath(which):
    """Returns the directory where the specified kind of data is
    stored.  System paths are assumed to be relative to the current
    working directory."""
    return os.path.abspath(which)


def getUserPath(which):
    """Returns the directory where the specified kind of data is
    stored for the current user.  The location of the user home
    directory depends on the operating system."""
    
    if customPaths.has_key(which):
        return customPaths[which]
    
    path = os.path.join(homeDir, which)

    # The system profiles directory and the user profiles directory can't
    # be the same.
    if which == PROFILES and path == getSystemPath(which):
        path = os.path.join(homeDir, 'user' + which)
    
    return path


def getBundlePaths(which):
    """Returns the directories where plugin bundles have their files
    of the specified kind."""
    return map(lambda p: os.path.join(p, which), bundlePaths)


def listPaths(which, userFirst=True):
    """Returns an array of search paths for the specified kind of files.
    """
    p = [getUserPath(which)] + getBundlePaths(which) + [getSystemPath(which)]
    if not userFirst:
        p.reverse()
    return p
               

def listFiles(which, userFirst=True):
    """Returns a list of all the files of the specified kind.

    @return An array of absolute file names.
    """
    files = []

    for path in listPaths(which, userFirst):
        try:
            for name in os.listdir(path):
                full = os.path.join(path, name)
                if full not in files:
                    files.append(os.path.abspath(full))
        except OSError:
            # The directory probably didn't exist.  We don't need to
            # do anything.
            pass

    files.sort()
    return files


def findBitmap(name):
    """Locates a bitmap file.  Files in the user's graphics directory
    override the contents of the system graphics directory.

    @param name Name of the bitmap without the file name extension.

    @return The full path to the bitmap file.
    """
    # First see if its in the user's graphics directory.
    for path in listPaths(GRAPHICS):
        for ext in ['jpg', 'bmp', 'png']:
            fileName = os.path.join(path, name + '.' + ext)
            if os.path.exists(fileName):
                return fileName

    # The file was not found.
    return ''


def getBase(path):
    """Returns the file name sans path and extension."""
    
    base = os.path.basename(path)
    pos = base.rfind('.')
    if pos < 0:
        # There is no extension.
        return base
    # Find the last dot.
    return base[:pos]


def hasExtension(extension, fileName):
    """Checks if the specified file name has the given extension.

    @param extension The extension to look for.

    @param fileName A file name.

    @return True, if the extension is found.
    """
    return re.search("^[^.#].*\." + extension + "$",
                     os.path.basename(fileName).lower()) != None


def quote(fileName):
    """Return the gives path in quotes so that it may be used as a
    command line option."""
    return '"' + fileName.replace('"', '""') + '"'


def fixSlashes(path):
    """Replace all forward and backward slashes with the correct
    slashes of the host OS."""  


def addAddonPath(path):
    """Define a new custom directory for addons.  Addons can be loaded
    from all the custom directories.  Installed addons always go to
    the default addons directory.

    @param path  Absolute path of the new custom addons directory.
    """
    if path not in addonPaths:
        addonPaths.append(path)


def removeAddonPath(path):
    """Remove a custom directory for addons.

    @param path  Absolute path of the custom addons directory.
    """
    if path in addonPaths:
        addonPaths.remove(path)
    else:
        logger.immediate(logger.HIGH, 'error-remove-addon-path-not-found', path)


def getAddonPaths():
    """Returns all the custom addon paths.

    @return An array containing full paths.
    """
    return [p for p in addonPaths]


def saveAddonPaths():
    """Write the custom addon paths to a configuration file."""

    fileName = os.path.join(getUserPath(CONF), 'addon-paths.conf')
    try:
        f = file(fileName, 'w')
        f.write('# This file is generated automatically.\n')
        f.write('configure addon-path (\n')
        for p in addonPaths:
            f.write('  readonly: %s\n' % p)
        f.write(')\n')
    except: 
        # Paths not saved.
        pass


def handleNotify(event):
    """When the <code>quit</code> notification is sent, save the
    custom addon paths."""

    if event.hasId('quit'):
        saveAddonPaths()


## When this module is initialized, check for the home directory.
## It is created automatically if it doesn't exist yet.
_checkSnowberryHome()

events.addNotifyListener(handleNotify, ['quit'])



# Parsing the current Doomsday version and release type from headers.

import os
import string

DOOMSDAY_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'doomsday')
    
DOOMSDAY_VERSION_FULL = "0.0.0-Name"
DOOMSDAY_VERSION_FULL_PLAIN = "0.0.0"
DOOMSDAY_VERSION_MAJOR = 0
DOOMSDAY_VERSION_MINOR = 0
DOOMSDAY_VERSION_REVISION = 0
DOOMSDAY_RELEASE_TYPE = "Unstable"

def parse_header_for_version(headerFile):
    versionMajor = 0
    versionMinor = 0
    versionRevision = 0
    versionName = ""
    releaseType = ""
    
    f = file(headerFile, 'rt')
    for line in f.readlines():
        line = line.strip()
        if line[:7] != "#define": continue
        if 'TEXTLONG' in line: continue
        baseAt = line.find("DOOMSDAY_VERSION_BASE")
        baseOff = 21
        if baseAt < 0:
            baseAt = line.find("GAME_VERSION_TEXT")
            baseOff = 17
        if baseAt < 0:
            baseAt = line.find("PLUGIN_VERSION_TEXT")
            baseOff = 20
        nameAt = line.find("DOOMSDAY_RELEASE_NAME")
        typeAt = line.find("DOOMSDAY_RELEASE_TYPE")
        if baseAt > 0:
            versionBase = line[baseAt + baseOff:].replace('\"','').strip()
            toks = versionBase.split('.')
            versionMajor = toks[0]
            versionMinor = toks[1]
            versionRevision = toks[2]
        if nameAt > 0:
            versionName = line[nameAt + 21:].replace('\"','').strip()
        if typeAt > 0:
            releaseType = line[typeAt + 21:].replace('\"','').strip()
            
    return (versionMajor, versionMinor, versionRevision, versionName, releaseType)

def find_version(quiet = False):
    if not quiet: print "Determining Doomsday version...",
    
    versionMajor, versionMinor, versionRevision, versionName, releaseType = \
        parse_header_for_version(os.path.join(DOOMSDAY_DIR, 'api', 'dd_version.h'))
    if not releaseType: releaseType = "Unstable"

    versionBase = "%s.%s.%s" % (versionMajor, versionMinor, versionRevision)

    global DOOMSDAY_VERSION_FULL
    global DOOMSDAY_VERSION_FULL_PLAIN
    global DOOMSDAY_VERSION_MAJOR
    global DOOMSDAY_VERSION_MINOR
    global DOOMSDAY_VERSION_REVISION
    global DOOMSDAY_RELEASE_TYPE
    
    DOOMSDAY_RELEASE_TYPE = releaseType
    DOOMSDAY_VERSION_FULL_PLAIN = versionBase
    DOOMSDAY_VERSION_FULL = versionBase
    if versionName:
        DOOMSDAY_VERSION_FULL += "-" + versionName    
    DOOMSDAY_VERSION_MAJOR = versionMajor
    DOOMSDAY_VERSION_MINOR = versionMinor
    DOOMSDAY_VERSION_REVISION = versionRevision

    if not quiet: print DOOMSDAY_VERSION_FULL + " (%s)" % releaseType


def version_summary(headerName):
    import sys
    import build_number
    buildNum = build_number.todays_build()

    # Get the Doomsday version. We can use some of this information.
    find_version(True)
    
    major, minor, revision, name, reltype = parse_header_for_version(headerName)
    if not reltype: reltype = DOOMSDAY_RELEASE_TYPE
    
    return "%s.%s.%s %s %s %s,%s,%s,%s" % (major, minor, revision, buildNum, reltype,
                                           major, minor, revision, buildNum)

        
# Invoked from qmake? Returns "version_base buildnum releasetype win32_version_with_buildnum"
if __name__ == '__main__':
    print version_summary(sys.argv[1])



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file cfparser.py Configuration File Parser
##
## This module implements a recursive descent parser that can be used
## for reading the various configuration files that Snowberry uses.
## The Parser offers the primary mechanism for parsing a file.
## Configuration files are composed of a number of hierarchically
## organized elements.

import string

class Element:
    """The base class for all Elements.  This class can't be used by
    itself.  Always use one of the subclasses."""
    
    def __init__(self, name):
        """Construct a new Element.  Element names are considered case
        independent."""
        self.name = name.lower()

    def isKey(self):
        """Returns true if this is a Key element."""
        return False

    def isList(self):
        """Returns true if this is a List element."""
        return False

    def isBlock(self):
        """Returns true if this is a Block element."""
        return False        
        
    def getName(self):
        """Returns the name of the element."""
        return self.name

    def setName(self, name):
        self.name = name

    def show(self, prefix=''):
        """Print the element's contents for debugging purposes."""
        print prefix + "Element(%s)" % self.getName()


class KeyElement(Element):
    "A KeyElement is an Element that contains a single value."
    
    def __init__(self, name, value):
        Element.__init__(self, name)
        self.value = value

    def isKey(self):
        "Returns true if this is a KeyElement."
        return True

    def getValue(self):
        "Return the value of the KeyElement."
        return self.value

    def getContents(self):
        """Similar behaviour as with a list element.  Returns the
        contents of the element as an array of strings.

        @return An array of strings containing only one element: the
        value of the element as a string.
        """
        return [self.getValue()]

    def show(self, prefix=''):
        "Print the element's contents for debugging purposes."
        print prefix + "KeyElement(%s, %s)" % (self.getName(), self.getValue())


class ListElement(Element):
    "A ListElement is an Element that contains a list of values."
    
    def __init__(self, name):
        """Constructor.
        
        @param name Name of the element.
        """
        Element.__init__(self, name)
        self.values = []
        
    def isList(self):
        "Returns true if this is a List element."
        return True

    def add(self, value):
        "Add a new value to the list."
        self.values.append(value)
        
    def getContents(self):
        """Returns the contents of the list element.
        
        @return An array of strings.
        """
        return self.values        

    def show(self, prefix=''):
        "Print the element's contents for debugging purposes."
        s = prefix + "ListElement(%s, [ " % self.getName()
        for v in self.values:
            s += "'%s' " % v
        s += "])"
        print s


class BlockElement(Element):
    "A BlockElement contains other Elements."
    
    def __init__(self, blockType, name):
        Element.__init__(self, name)
        self.type = blockType.lower()
        self.contents = []

    def isBlock(self):
        "Returns true if this is a Block element."
        return True   

    def add(self, childElement):
        "Add a new element to this block's contents."
        if not childElement:
            raise Exception("attempted to append an invalid child element to a block")

        self.contents.append(childElement)

    def show(self, prefix=''):
        "Print the element's contents for debugging purposes."
        print prefix + "BlockElement(type=%s, name=%s," % (self.getType(),
                                                           self.getName())
        subPrefix = prefix + 3 * ' '
        for c in self.contents:
            c.show(subPrefix)
        print prefix + ")"

    def getType(self):
        "Returns the type identifier of the block."
        return self.type

    def getContents(self):
        "Returns the contents array of this block."
        return self.contents

    def find(self, elementName):
        """Returns the element with the specified name.  If it doesn't
        exist, returns None."""

        # Element names are considered case independent.
        elementName.lower()
        
        for e in self.contents:
            if e.getName() == elementName:
                return e

        # The specified name was not in this block's contents.
        return None

    def findValue(self, elementName):
        """Returns the value of the KeyElement with the specified
        name.  If it doesn't exist, returns an empty string."""
        e = self.find(elementName)
        if not e:
            return ''
        else:
            return e.getValue()


class ParseFailed(Exception):
    """An exception that is raised when the Parser encounters a syntax
    error in the source file."""

    def __init__(self, reason=''):
        Exception.__init__(self, reason)


class OutOfElements(Exception):
    """This exception is thrown when there are no more
    elements to return from Parser.get()."""

    def __init__(self):
        Exception.__init__(self, "no more elements in source file")


class Parser:
    """The Parser class is used for reading the contents of
    configuration, profile, and metadata files."""

    ## Characters that cause a token to end.
    TOKEN_BREAKING_CHARS = '#:=()<>,"' + string.whitespace
    
    def __init__(self, content):
        """Initialize the parser for reading a block of source content.

        @param content The configuration data that is to be parsed.
        """
        # The source data.  Add an extra newline so the character
        # reader won't get confused.
        self.content = content + "\n"
        # Index of the next character from the source.
        self.cursor = 0
        
        self.nextChar()
        self.tokenStartOffset = 0

        # When nextToken() is called and the current token is empty,
        # it is deduced that the source file has ended. We must
        # therefore set a dummy token that will be discarded
        # immediately.
        self.currentToken = ' '

        self.nextToken()

    def peekChar(self):
        "Returns the next character from the source file."
        return self.currentChar

    def nextChar(self):
        "Move to the next character in the source file."
        if self.cursor >= len(self.content):
            # No more characters to read.
            raise EOFError

        self.currentChar = self.content[self.cursor]
        self.cursor += 1

    def readLine(self):
        """Read a line of text from the content and return it."""
        
        line = ""

        self.nextChar()
        while self.currentChar != '\n':
            line += self.currentChar
            self.nextChar()

        return line        

    def readToEOL(self):
        """Read until a newline is encountered.  Returns the contents
        of the line."""
        self.cursor = self.tokenStartOffset
        line = self.readLine()
        try:
            self.nextChar()
        except EOFError:
            # If the file ends right after the line, we'll get the EOF
            # exception.  We can safely ignore it for now.
            pass
        return line
        
    def peekToken(self):
        return self.currentToken

    def nextToken(self):
        "Returns the next meaningful token from the source file."

        # Already drawn a blank?
        if self.currentToken == "":
            raise EOFError("out of tokens")

        self.currentToken = ""

        try:
            # Skip over any whitespace.
            while self.peekChar() in string.whitespace + '#':
                # Comments are considered whitespace.
                if self.peekChar() == '#':
                    self.readLine()
                self.nextChar()

            # Store the offset where the token begins.
            self.tokenStartOffset = self.cursor

            # The first nonwhite is accepted.
            self.currentToken += self.peekChar()
            self.nextChar()

            # Token breakers are tokens all by themselves.
            if self.currentToken[0] in Parser.TOKEN_BREAKING_CHARS:
                return self.currentToken
            
            while self.peekChar() not in Parser.TOKEN_BREAKING_CHARS:
                self.currentToken += self.peekChar()
                self.nextChar()

        except EOFError:
            # If the source file ends, return an empty token.
            pass

        return self.currentToken

    def get(self):
        """This is the method that the user calls to retrieve the next
        element from the source file.  If there are no more elements
        to return, a OutOfElements exception is raised."""
        e = self.parseElement()
        if not e: raise OutOfElements
        return e

    def parseElement(self):
        """Returns the next element from the source file.

        @return An instance of one of the Element classes.
        """
        
        try:
            key = self.peekToken()

            # The next token decides what kind of element we have here.
            next = self.nextToken()
            
        except EOFError:
            # The file ended.
            return None

        if next == ':' or next == '=':
            return self.parseKeyElement(key)
        elif next == '<':
            return self.parseListElement(key)
        else:
            # It's a block element.
            return self.parseBlockElement(key)

    def parseString(self):
        """Parse a string literal.  Returns the string sans the
        quotation marks in the beginning and the end."""

        if self.peekToken() != '"':
            raise ParseFailed('expected string to begin with \'"\', but \'%s\' found instead' % self.peekToken())

        # The collected characters.
        chars = []

        while self.peekChar() != '"':
            if self.peekChar() == "'":
                # Double single quotes form a double quote ('' => ").
                self.nextChar()
                if self.peekChar() == "'":
                    chars.append('"')
                else:
                    chars.append("'" + self.peekChar())
            else:
                # Other characters are appended as-is, even newlines.
                chars.append(self.peekChar())
            self.nextChar()

        # Move the parser to the next token.
        self.nextChar()
        self.nextToken()

        # Compile a string out of all the characters.
        return string.join(chars, '')

    def parseValue(self):
        """Parse a value from the source file.  The current token
        should be on the first token of the value.  Values come in
        different flavours:
        - single token
        - string literal (can be split)"""

        # Check if it is the beginning of a string literal.
        if self.peekToken() == '"':
            value = ""
            try:
                # The value will be composed of any number of
                # sub-strings.
                while True:
                    value += self.parseString()
            except:
                # No more strings to append.
                return value

        # Then it must be a single token.
        value = self.peekToken()
        self.nextToken()
        return value

    def parseKeyElement(self, name):
        """Parse a key element.
        @param name Name of the parsed key element."""

        # A colon means that that the rest of the line is the value of
        # the key element.
        if self.peekToken() == ':':
            value = self.readToEOL().strip()
            self.nextToken()
        else:
            #
            # Key =
            #   "This is a long string "
            #   "that spans multiple lines."
            #
            self.nextToken()
            value = self.parseValue()

        return KeyElement(name, value)
               
    def parseBlockElement(self, blockType):
        """Parse a block element, identified by the given name.

        @param blockType Identifier of the block.
        """
        
        block = BlockElement(blockType, self.peekToken())

        # How about some attributes?
        # Syntax: {token value} '('
        
        try:
            self.nextToken()
            while self.peekToken() != '(':
                keyName = self.peekToken()
                self.nextToken()
                value = self.parseValue()

                # This becomes a key element inside the block.
                block.add(KeyElement(keyName, value))

            # Move past the opening parentheses.
            self.nextToken()

            while self.peekToken() != ')':
                element = self.parseElement()
                if element == None:
                    raise ParseFailed("block element was never closed, end of file encountered before ')' was found")
                block.add(element)

        except EOFError:
            raise ParseFailed("end of file encountered unexpectedly while parsing a block element")

        # Move past the closing parentheses.
        self.nextToken()

        return block

    def parseListElement(self, name):
        "Parse a list element, identified by the given name."
        element = ListElement(name)

        if self.peekToken() != '<':
            raise ParseFailed("list must begin with a '<', but '%s' found instead" % self.peekToken())

        # List syntax:
        # list ::= list-identifier '<' [value {',' value}] '>'
        # list-identifier ::= token

        # Move past the opening angle bracket.
        self.nextToken()

        while True:
            element.add(self.parseValue())

            # List elements are separated explicitly.
            separator = self.peekToken()
            self.nextToken()

            # The closing bracket?
            if separator == '>':
                break

            # There should be a comma here.
            if separator != ',':
                raise ParseFailed("list values must be separated with a comma, but '%s' found instead" % separator)

        return element


class FileParser (Parser):
    """A parser for reading configuration data directly from a file."""

    def __init__(self, fileName):
        """Construct a FileParser.

        @param fileName Path of the source file.
        """
        self.contentFile = file(fileName)
        Parser.__init__(self, self.contentFile.read())



import sys, os, subprocess

isDryRun = '--dry-run' in sys.argv

sys.path += ['/Users/jaakko/Dropbox/Scripts']
import dew

comment     = 'Generated from Amethyst source by wikidocs.py'
collections = ['engine', 'libcommon', 'libdoom', 'libheretic', 'libhexen']
modes       = ['command', 'variable']


def categoryForMode(m):
    """Wiki category for the mode."""
    return {'command':  'Console command', 
            'variable': 'Console variable'}[m]

def categoryForCollection(col, m):
    """Wiki category for a collection."""
    if col == 'engine': col = 'Engine'
    return categoryForMode(m) + ' (%s)' % col
    
def heading(col, m):
    """Heading on the index page for a particular collection."""
    hd = {'command': 'Commands', 'variable': 'Variables'}[m]
    if col == 'engine': col = 'Doomsday'
    elif col == 'libcommon': col = 'all games'
    hd += ' for '  + col
    return hd

def colWidth(m):
    """Width of an index table column."""
    if m == 'variable': return 33
    return 20

def colHeading(m):
    """Index table heading."""
    hd = {'command': 'Command', 'variable': 'Variable'}[m]
    return hd
    
def amethyst(input):
    """Runs amethyst with the given input and returns the output."""
    p = subprocess.Popen(['amethyst', '-dWIKI'], 
        stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (output, errs) = p.communicate(input)
    return output    

def inOtherCollection(title, excludeCol):
    """Check if a @a title exists in any other collection other
    than @a excludeCol."""
    for col in collections:
        if col == excludeCol: continue
        if title in pagesForCollection[col]:
            return True
    return False
    
    
class Page:
    """Wiki article about a console variable or command."""
    
    def __init__(self, title, content):
        self.title = title
        self.content = content
        self.name = ''
        self.summary = ''
        self.mode = ''
        
    def appendCollectionToTitle(self, col):
        if ')' in self.title:
            self.title = self.title.replace(')', ' in %s)' % col)
        else:
            self.title += ' (%s)' % col
        

pagesForCollection = {}

for col in collections:
    print 'Processing', col, 'documentation...'
    pagesForCollection[col] = {}
            
    for mode in modes:
        print '-', mode

        dirName = os.path.join(col, mode)
        files = os.listdir(dirName)
        for fn in files:
            if not fn.endswith('.ame'): continue
            name = fn[:-4]
            title = name

            if mode == 'command': title += " (Cmd)" 

            # Generate full page.            
            templ = '@require{amestd}\n'
            templ += '@macro{summary}{== Summary == @break @arg}\n'
            templ += '@macro{description}{@break == Description == @break @arg}\n'
            templ += '@macro{cbr}{<br/>}\n'
            templ += '@macro{usage}{@break === Usage === @break @arg}\n'
            templ += '@macro{examples}{@break === Examples === @break @arg}\n'
            templ += '@macro{seealso}{@break === See also === @break @arg}\n'
            templ += '@macro{params}{@break === Usage === @break @arg}\n'
            templ += '@macro{example}{@break === Example === @break @arg}\n'
            templ += '@begin\n' + \
                 file(os.path.join(dirName, fn)).read()
                    
            templ += '\n[[Category:%s]] @br\n' % categoryForMode(mode)
            templ += '[[Category:%s]]\n' % categoryForCollection(col, mode)
                      
            page = Page(title, amethyst(templ))
            page.mode = mode
            pagesForCollection[col][title] = page
            
            if isDryRun:
                print page.title
                print page.content
            
            # Just the summary for the index.
            templ = '@require{amestd}\n'
            templ += '@macro{summary}{@arg}\n'
            templ += '@macro{description}{}\n'
            templ += '@macro{cbr}{}\n'
            templ += '@begin\n' + \
                 file(os.path.join(dirName, fn)).read()
                 
            page.name = name
            page.summary = amethyst(templ).strip()
                                        
# Check for ambiguous pages.
ambigs = []
for col in collections:
    pages = pagesForCollection[col]
    for page in pages.values():
        if inOtherCollection(page.title, col):
            title = page.title
            ambigPage = Page(title, 
                "'''%s''' can be one of the following:\n" % title)
            # Must be more specific.
            for c in collections:
                if title in pagesForCollection[c]:
                    p = pagesForCollection[c][title]
                    p.appendCollectionToTitle(c)
                    ambigPage.content += '* [[%s]]\n' % p.title
            ambigs.append(ambigPage)
            
ambigs = sorted(ambigs, key=lambda p: p.title)
        
# Generate the indices.
indexPage = {}
indexPage['variable'] = ''
indexPage['command'] = ''

for col in collections:
    for mode in modes:
        indexPage[mode] += '\n== %s ==\n\n' % heading(col, mode)
        indexPage[mode] += '{| class="wikitable" width="100%%"\n! width="%i%%"| %s\n! Description\n' \
            % (colWidth(mode), colHeading(mode))
        
        pages = sorted(pagesForCollection[col].values(), key=lambda p: p.name)
        for page in filter(lambda p: p.mode == mode, pages):
            indexPage[mode] += '|-\n| [[%s|%s]]\n| %s\n' % (page.title, page.name, page.summary)

        indexPage[mode] += '|}\n'
        
for m in modes:
    indexPage[m] += '\n[[Category:Console]]\n[[Category:References]]\n'

if isDryRun:
    print indexPage['command']
    print indexPage['variable']
    for amb in ambigs:
        print amb.title
        print amb.content
    sys.exit()   
    
        
dew.login()
        
dew.submitPage('Console command reference', indexPage['command'], comment)
dew.submitPage('Console variable reference', indexPage['variable'], comment)

print 'Submitting deambiguation pages...'

for amb in ambigs:
    print '-', amb.title
    dew.submitPage(amb.title, amb.content, comment)

print 'Submitting pages...'

for col in collections:
    pages = pagesForCollection[col]
    count = len(pages)
    i = 0
    for page in pages.values():
        i += 1
        print '-', col, ": %i / %i" % (i, count)
        dew.submitPage(page.title, page.content, comment)

dew.logout()


#!/usr/local/bin/python
"""
EXPERIMENTAL CODE!

Here we load a .TTF font file, and display it in
a basic pygame window. It demonstrates several of the
Font object attributes. Nothing exciting in here, but
it makes a great example for basic window, event, and
font management.
"""

import sys, struct
import pygame
import math
from pygame.locals import *
from pygame import Surface
from pygame.surfarray import blit_array, make_surface, pixels3d, pixels2d
import Numeric

from Foundation import *
from AppKit import *


# Version 1 is the monochrome one.
DFN_VERSION = 2


def _getColor(color=None):
    if color is None:
        return NSColor.clearColor()
    return NSColor.colorWithDeviceRed_green_blue_alpha_(*color)
    #div255 = (0.00390625).__mul__
    #if len(color) == 3:
    #    color = tuple(color) + (255.0,)
    #return NSColor.colorWithDeviceRed_green_blue_alpha_(*color) #*map(div255, color))


def _getBitmapImageRep(size, hasAlpha=True):
    width, height = map(int, size)
    return NSBitmapImageRep.alloc().initWithBitmapDataPlanes_pixelsWide_pixelsHigh_bitsPerSample_samplesPerPixel_hasAlpha_isPlanar_colorSpaceName_bytesPerRow_bitsPerPixel_(None, width, height, 8, 4, hasAlpha, False, NSDeviceRGBColorSpace, width*4, 32)


def alpha_blend(src, dest):
    a = src[3]/255.0
    inv = 1.0  - a
    #return src
    #if src[3] == 0:
    #return src
    #print src[3], a, inv, int(a * src[3] + inv * dest[3])
    return ( min(255, int(a * 1 * src[0] + inv * dest[0])),
             min(255, int(a * 1 * src[1] + inv * dest[1])),
             min(255, int(a * 1 * src[2] + inv * dest[2])),
             min(255, int(a * src[3] + inv * dest[3])) )


def filter_range(n):
    offset = 1 # - int(n/5)
    return range(-n/2 + offset, n/2 + offset)


class SysFont(object):
    def __init__(self, name, size, variant=None):
        self._font = NSFont.fontWithName_size_(name, size)
        self._isBold = False
        self._isOblique = False
        self._isUnderline = False
        self._variant = variant
        self._family = name
        self._size = size
        self._setupFont()

    def _setupFont(self):
        name = self._family
        if self._variant:
            name = '%s-%s' % (name, self._variant)
        elif self._isBold or self._isOblique:
            name = '%s-%s%s' % (
                name,
                self._isBold and 'Bold' or '',
                self._isOblique and 'Oblique' or '')
        self._font = NSFont.fontWithName_size_(name, self._size)
        print name, self._font
        if self._font is None:
            if self._isBold:
                self._font = NSFont.boldSystemFontOfSize(self._size)
            else:
                self._font = NSFont.systemFontOfSize_(self._size)
        
    def get_ascent(self):
        return self._font.ascender()

    def get_descent(self):
        return -self._font.descender()

    def get_bold(self):
        return self._isBold

    def get_height(self):
        return self._font.defaultLineHeightForFont()

    def get_italic(self):
        return self._isOblique

    def get_linesize(self):
        pass

    def get_underline(self):
        return self._isUnderline

    def set_bold(self, isBold):
        if isBold != self._isBold:
            self._isBold = isBold
            self._setupFont()

    def set_italic(self, isOblique):
        if isOblique != self._isOblique:
            self._isOblique = isOblique
            self._setupFont()

    def set_underline(self, isUnderline):
        self._isUnderline = isUnderline
        
    def size(self, text):
        return tuple(map(int,map(math.ceil, NSString.sizeWithAttributes_(text, {
            NSFontAttributeName: self._font,
            NSUnderlineStyleAttributeName: self._isUnderline and 1.0 or 0,
        }))))
        
    def advancement(self, glyph):
        return tuple(self._font.advancementForGlyph_(ord(glyph)))

    def shadow_filter_size(self):
        filter_size = int(self.get_height() / 5)
        if filter_size % 2 == 0:
            filter_size += 1
        return int(max(filter_size, 3))

    def border_size(self):
        return self.shadow_filter_size()/2
    
    def render(self, text, antialias = True, 
               forecolor=(255, 255, 255, 255), 
               backcolor=(255, 255, 255, 0)):
        size = self.size(text)
        img = NSImage.alloc().initWithSize_(size)
        img.lockFocus()

        NSString.drawAtPoint_withAttributes_(text, (0.0, 0.0), {
            NSFontAttributeName: self._font,
            NSUnderlineStyleAttributeName: self._isUnderline and 1.0 or 0,
            NSBackgroundColorAttributeName: _getColor(backcolor),# or None,
            NSForegroundColorAttributeName: _getColor(forecolor),
        })

        rep = NSBitmapImageRep.alloc().initWithFocusedViewRect_(((0.0, 0.0), size))
        img.unlockFocus()
        if rep.samplesPerPixel() == 4:
            s = Surface(size, SRCALPHA|SWSURFACE, 32, [-1<<24,0xff<<16,0xff<<8,0xff])
            
            a = Numeric.reshape(Numeric.fromstring(rep.bitmapData(), typecode=Numeric.Int32), (size[1], size[0]))
            blit_array(s, Numeric.swapaxes(a, 0, 1))
            #print "surface size is ", size

            filter_size = self.shadow_filter_size()
            border = self.border_size()
            
            result = []
            result_size = (size[0] + border*2, size[1] + border*2)
            #print "result size is ", result_size
            for y in range(result_size[1]):
                result.append([(0, 0, 0, 0)] * result_size[0])
            
            # Filter the character appropriately.
            s.lock()
            for y in range(size[1]):
                for x in range(size[0]):
                    color = s.get_at((x, y))
                    s.set_at((x, y), (255, 255, 255, color[3]))

            if filter_size == 3:
                factors = [[1, 2, 1],
                           [2, 4, 2],
                           [1, 2, 1]]
            elif filter_size == 5:
                factors = [[1,  4,  6,  4,  1],
                           [4,  16, 24, 16, 4],
                           [6,  24, 36, 24, 6],
                           [4,  16, 24, 16, 4],
                           [1,  4,  6,  4,  1]]
            elif filter_size == 7:
                factors = [[1,  6,   15,  20,  15,  6,   1],
                           [6,  36,  90,  120, 90,  36,  6],
                           [15, 90,  225, 300, 225, 90,  15],
                           [20, 120, 300, 400, 300, 120, 20],
                           [15, 90,  225, 300, 225, 90,  15],
                           [6,  36,  90,  120, 90,  36,  6],
                           [1,  6,   15,  20,  15,  6,   1]]
            else:
                print "factors for filter size", filter_size, "not defined!"
                factors = 0

            # Filter a shadow.
            for ry in range(result_size[1]):
                for rx in range(result_size[0]):
                    inpos = (rx - int(border), ry - int(border))
                    count = 0
                    colorsum = 0
                    u = 0
                    v = 0
                    for a in filter_range(filter_size):
                        v = 0
                        for b in filter_range(filter_size):
                            x, y = (inpos[0] + a, inpos[1] + b)
                            factor = factors[u][v]
                            if x >= 0 and y >= 0 and x < size[0] and y < size[1]:
                                colorsum += factor * s.get_at((x, y))[3]
                            count += factor
                            v += 1
                        u += 1
                    if count > 0:
                        result_color = (int(colorsum) + count/2) / count
                    else:
                        result_color = 0
                    result[ry][rx] = (0, 0, 0, min(255, result_color*1.4))

            # Blend the glyph itself to the result.
            for y in range(size[1]):
                for x in range(size[0]):
                    color = s.get_at((x, y))
                    result[border + y][border + x] = \
                        alpha_blend(color, result[border + y][border + x])
                    result[border + y][border + x] = \
                        alpha_blend(color, result[border + y][border + x])
                    #result[border + y][border + x] = \
                    #    alpha_blend(color, result[border + y][border + x])

            s.unlock()
                    
            glyph_rgba = Numeric.array(result, typecode=Numeric.UnsignedInt8)
            return glyph_rgba, result_size
                
            #return s.convert_alpha()


class Glyph:
    def __init__(self, code):
        self.code = code
        self.bitmap = None
        self.pos = (0, 0)
        self.dims = (0, 0)


def pow2(n):
    i = 1
    while i < n: 
        i *= 2
    return i
    
    
def find_pixel(glyphs, x, y):
    for g in glyphs:
        lx = x - g.pos[0]
        ly = y - g.pos[1]
        if lx >= 0 and ly >= 0 and lx < g.dims[0] and ly < g.dims[1]:
            # This is the glyph.
            return g.bitmap[ly][lx]
    return (0, 0, 0, 0)


if __name__=='__main__':
    if len(sys.argv) < 5:
        print "usage: " + sys.argv[0] + " (font-name) (font-size) (texture-width) (output-file)"
        sys.exit(0)

    out_filename = sys.argv[4]
   
    pygame.init()
    #screen = pygame.display.set_mode((600, 600))
    point_size = int(sys.argv[2])
    s = SysFont(sys.argv[1], point_size)
    #s = SysFont('Andale Mono', 24)
    #done = False

    # Collect all the glyphs here.
    glyphs = []
    
    texture_width = int(sys.argv[3])
    font_point_height = s.get_height()
    print "Font texture width =", texture_width
    print "Line height =", font_point_height, "pt"
    print "Font size =", point_size, "pt"
    
    ascent = s.get_ascent()
    descent = s.get_descent()
    print "Ascent =", ascent, "pt  Descent =", descent, "pt"
    
    print "Shadow filter is sized %i for this font." % s.shadow_filter_size()
    border_size = s.border_size()
    print "Glyph borders are %i pixels wide." % border_size
    
    rover = (0, 0)
    line_height = 0
    texture_height = 0
    used_pixels = 0
    
    # Only the basic ASCII set, and some specific characters.
    char_range = range(0x20, 0x7F) + [0x96, 0x97, 0xA3, 0xA4, 0xA9, 0xB0, 0xB1]
    for code in char_range: 
        glyph = Glyph(code)
        try:
            glyph.bitmap, glyph.dims = s.render(chr(code))
        
            used_pixels += glyph.dims[0] * glyph.dims[1]
        
            line_height = max(line_height, glyph.dims[1])
            texture_height = max(line_height, texture_height)
        
            # Does if fit here?
            if rover[0] + glyph.dims[0] > texture_width:
                # No, move to the new row.
                rover = (0, rover[1] + line_height)
                texture_height += line_height
                print "Finished a %i pixels tall line." % line_height
                line_height = 0
            
            glyph.pos = rover
        
            # Move the rover.
            rover = (rover[0] + glyph.dims[0], rover[1])
        
            #print "Code %i: Placement =" % code, glyph.pos, " dims =", glyph.dims
            glyphs.append(glyph)
        #except Exception, x:
        except KeyboardInterrupt:
            print "Code %i: Failed to render." % code
            print x
        
    print "Finished rendering %i glyphs." % len(glyphs)
    actual_line_height = glyphs[0].dims[1] - border_size*2
    print "Actual line height is %i pixels." % actual_line_height
    
    print "Size of the finished texture is %i x %i." % (texture_width, texture_height)
        
    actual_height = pow2(texture_height)
    total_pixels = texture_width * actual_height
    
    print "Unused texture area: %.2f %%" % (100*(1.0 - used_pixels/float(total_pixels)))
    print "Writing output to file: %s" % out_filename
    
    out = file(out_filename, 'wb')
    
    # The DFN format begins with a single byte that declares the version 
    # of the format.
    out.write(struct.pack('B', DFN_VERSION))
    
    # The header follows:
    # Pixel format (0 = RGBA, 1 = LA).
    out.write(struct.pack('B', 0)) 
    # Glyph count, texture width and height (unsigned shorts).
    out.write(struct.pack('<HHH', texture_width, texture_height, len(glyphs)))
    # Border width and line height.
    out.write(struct.pack('<HH', border_size, actual_line_height))

    # Scaling factor for converting points to pixels.
    scale = float(actual_line_height) / font_point_height
    pixel_ascent = int(math.ceil(scale * ascent))
    pixel_descent = int(math.ceil(scale * descent))
    print "    Pixel ascent/descent:", pixel_ascent, '/', pixel_descent
    
    pixel_line_height = int(scale * font_point_height)
    print "    Pixel line height:", pixel_line_height
    
    # Ascent and descent.
    out.write(struct.pack('<HHH', pixel_line_height, pixel_ascent, pixel_descent))
    
    # Glyph metrics.
    for g in glyphs:
        out.write(struct.pack('<HHHHH', g.code, g.pos[0], g.pos[1], 
                              g.dims[0], g.dims[1]))

    # Glyph map.
    print "    Writing glyph map..."
    for y in range(texture_height):
        for x in range(texture_width):
            pixel = find_pixel(glyphs, x, y)
            # Use the LA format.
            #out.write(struct.pack('BB', pixel[0], pixel[3]))
            # Use the RGBA format.
            out.write(struct.pack('BBBB', pixel[0], pixel[1], pixel[2], pixel[3]))
    
    out.close()
    


# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file area.py Area Class
##
## Each area can be divided into further sub-areas, which are treated
## the same way as the top-level areas.  The Area class can represent
## either a top-level area or a sub-area.

import sys, os, wx, string
import wx.wizard as wiz
import host, events, widgets, language, paths, ui
import base
import sb.widget.bar
import sb.widget.base
import sb.widget.button
import sb.widget.field
import sb.widget.list
import sb.widget.tab
import sb.widget.text
import sb.widget.tree
import sb.profdb as pr
import logger
from widgets import uniConv

# TODO: Make these .conf'able.
SETTING_WEIGHT_LEFT = 1
SETTING_WEIGHT_RIGHT = 2

# Border widths in areas.
# TODO: Move these to a .conf.
if host.isWindows():
    AREA_BORDER = 2
    AREA_BORDER_BOXED = 2
elif host.isUnix():
    AREA_BORDER = 2
    AREA_BORDER_BOXED = 1
else:
    AREA_BORDER = 3
    AREA_BORDER_BOXED = 2

# wxWidget ID counters.
WIDGET_ID = 100


class Area (base.Widget):
    """Each Area instance governs the contents of a wxPanel widget.
    Together with the classes in widgets.py, the Area class
    encapsulates all management of the user interface.  Modules that
    use ui.py or widgets.py need not access wxWidgets directly at all.

    The Area class implements the base.Widget interface from
    widgets.py.  getWxWidget() and clear() are the important
    Widget-inherited methods that Area uses.

    Area alignment constants:
    - ALIGN_HORIZONTAL: Widgets are placed next to each other
      inside the area.
    - ALIGN_VERTICAL: New widgets are placed below the existing
      widgets.
    """

    def __init__(self, id, panel, alignment=0, border=0, parentArea=None,
                 borderDirs=15):
        """The constructor initializes the Area instance so that it
        can be used immediately afterwards.

        @param panel The wxWidget panel inside which the new Area will
        be.

        @param alignment Alignment for widgets inside the area.  This
        can be either ALIGN_VERTICAL or ALIGN_HORIZONTAL.

        @param border Width of the border around the new Area.

        @param parentArea Area instance that will act as the parent of
        the new area.  If this is None, it is assumed that the new
        Area is directly inside the main panel of the dialog.
        """
        self.id = id
        self.panel = panel
        self.border = border
        self.borderDirs = borderDirs

        # Widget layout parameters.
        # Weight for sizer when adding widgets.
        self.weight = 1
        self.expanding = True

        # Create a Sizer that maintains the proper layout of widgets
        # inside the Area.
        if alignment == ui.ALIGN_HORIZONTAL:
            self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        else:
            self.sizer = wx.BoxSizer(wx.VERTICAL)

        # Top-level areas have a wx.Panel, with which the sizer needs
        # to be associated.  Subareas use the parent's area.
        if not parentArea:
            self.panel.SetSizer(self.sizer)
            self.parentArea = None
        else:
            self.parentArea = parentArea

        # The contents of the area will be added inside this sizer.
        self.containerSizer = self.sizer

        # The arrays that contain information about the contents of
        # the Area.
        self.widgets = []

    def setId(self, ident):
        self.id = ident

    def getId(self):
        return self.id

    def getPanel(self):
        return self.panel

    def getWxWidget(self):
        """An Area instance can be used like a Widget."""
        return self.sizer

    def __getNewId(self):
        """Generate a new widget ID number.  These are given to
        wxWidgets so the events they send can be handled.
        """
        global WIDGET_ID
        id = WIDGET_ID
        WIDGET_ID += 1
        return id

    def setWeight(self, weight):
        """Set the sizer weight that will be used for the next widget
        that is created."""
        self.weight = weight

    def setBorder(self, border, dirs=None):
        """Set the width of empty space around widgets.

        @param border Number of pixels for new borders.
        """
        self.border = border
        if dirs is not None:
            self.borderDirs = dirs

    def setBorderDirs(self, dirs):
        self.borderDirs = dirs

    def setBackgroundColor(self, red, green, blue):
        self.panel.SetBackgroundColour(wx.Colour(red, green, blue))
        self.panel.SetBackgroundStyle(wx.BG_STYLE_COLOUR)

    def setExpanding(self, expanding):
        self.expanding = expanding

    def updateLayout(self, redoLayout=True):
        """Recalculate the layout of all the widgets inside the area.
        This must be called when widgets are created or destroyed at
        runtime.  The layout of all areas is updated automatically
        right before the main window is displayed during startup.  If
        a plugin modifies the layout of its widgets during event
        handling, it must call updateLayout() manually.
        """
        # Tell the main sizer of our panel's current minimum size.
        m = self.sizer.GetMinSize()
        topSizer = self.panel.GetContainingSizer()
        if topSizer:
            topSizer.SetItemMinSize(self.panel, m)
        # Redo the layout.
        if redoLayout:
            self.sizer.Layout()

    def addSpacer(self):
        """Adds empty space into the area.  The size of the empty
        space is determined by the current weight.
        """
        if self.expanding:
            self.containerSizer.AddStretchSpacer(self.weight)
        else:
            self.containerSizer.Add((0, 0), self.weight, self._getLayoutFlags(),
                                    self.border)
        pass

    def __addWidget(self, widget):
        """Insert a new widget into the area.
        @param wxWidget Widget object to add to the area.
        """
        # The Area keeps track of its contents.
        self.widgets.append(widget)

        # Use the low-level helper to do the adding.
        self._addWxWidget(widget.getWxWidget())

    def _getLayoutFlags(self):
        """Determines the appropriate layout flags for child widgets."""
        flags = 0
        if self.expanding: flags |= wx.EXPAND
        if self.border > 0:
            if self.borderDirs & 0x1:
                flags |= wx.TOP
            if self.borderDirs & 0x2:
                flags |= wx.RIGHT
            if self.borderDirs & 0x4:
                flags |= wx.BOTTOM
            if self.borderDirs & 0x8:
                flags |= wx.LEFT
        flags |= wx.ALIGN_CENTER_VERTICAL
        flags |= wx.ALIGN_CENTER
        return flags

    def _addWxWidget(self, wxWidget):
        """Insert a wxWidget into the area's sizer.  This protected
        low-level helper method takes care of selecting the correct
        layout parameters for the widget.

        @param wxWidget A wxWidget object.
        """
        # Insert the widget into the Sizer with appropriate layout
        # parameters.
        self.containerSizer.Add(wxWidget, self.weight,
                                self._getLayoutFlags(), self.border)

    def destroyWidget(self, widget):
        """Destroy a particular widget inside the area.

        @param widget The widget to destroy.
        """
        if widget in self.widgets:
            widget.clear()
            widget.destroy()
            self.widgets.remove(widget)

    def destroy(self):
        """Destroy the area and its contents."""

        self.clear()

        if not self.parentArea:
            # Standalone areas are never added to a sizer.
            return

        # Detach and destroy the sizer of this area.
        self.parentArea.getWxWidget().Detach(self.getWxWidget())
        self.getWxWidget().Destroy()

    def clear(self):
        """Remove all the widgets inside the area."""

        # Remove each widget individually.
        for widget in self.widgets:

            # Tell the widget it is about to the removed.  Subareas
            # will process their contents recursively.
            widget.clear()

            # The widget must be explicitly destroyed.
            widget.destroy()

        # No widgets remain.
        self.widgets = []

    def enable(self, doEnable):
        """Enabling an area means enabling all the widgets in it."""
        for widget in self.widgets:
            widget.enable(doEnable)

    def createChildPanel(self):
        """Create a new wxPanel inside this area's panel.  The new
        panel can be given to a subarea, whose background color can
        then be set."""

        return wx.Panel(self.panel, -1)

    def createArea(self, alignment=1, border=0, boxedWithTitle=None,
                   backgroundColor=None):
        """Create a sub-area that is a part of the parent area and can
        contain widgets.

        @param alignment  Widget alignment inside the sub-area.  Can be
        either ALIGN_VERTICAL or ALIGN_HORIZONTAL.

        @param border  Border width for the sub-area (pixels).

        @param boxedWithTitle  If you want to enclose the area inside a
        box frame, give the identifier of the box title here.  The
        actual label text is the translation of the identifier.

        @param backgroundColor  If specified, the new area will have
        its own wxPanel with the color as a background color.

        @return An Area object that represents the contents of the
        area.
        """
        # Subareas don't have identifiers.
        if boxedWithTitle:
            subArea = BoxedArea(boxedWithTitle, self, self.panel,
                                alignment, border)
        elif backgroundColor:
            subArea = ColoredArea(backgroundColor, self, self.panel,
                                  alignment, border)
        else:
            subArea = Area('', self.panel, alignment, border, self)

        self.__addWidget(subArea)
        return subArea

    def createMultiArea(self):
        multi = MultiArea('', self)
        self.__addWidget(multi)
        return multi

    def createButton(self, name, style=sb.widget.button.Button.STYLE_NORMAL):
        """Create a button widget inside the area.

        @param name Identifier of the button.  The text label that
        appears in the button will be the translation of this
        identifier.

        @return A sb.widget.button.Button object.
        """
        w = sb.widget.button.Button(self.panel, self.__getNewId(), name, style)
        self.__addWidget(w)
        return w

    def createLine(self):
        """Create a static horizontal divider line."""

        widget = widgets.Line(self.panel)
        self.__addWidget(widget)
        return widget

    def createImage(self, imageFile):
        """Create an image widget inside the area.

        @param imageFile Base name of the image file.
        """
        widget = widgets.Image(self.panel, imageFile)
        self.__addWidget(widget)
        return widget

    def createText(self, name='', suffix='', maxLineLength=0,
                   align=sb.widget.text.Text.LEFT):
        """Create a text label widget inside the area.

        @param name Identifier of the text label widget.  The actual
        text that appears in the widget is the translation of this
        identifier.

        @return A widgets.Text object.
        """
        widget = sb.widget.text.Text(self.panel, -1, name, suffix,
                                     maxLineLength, align)
        self.__addWidget(widget)
        return widget

    def createTextField(self, name=''):
        """Create a text field widget inside the area.

        @param name Identifier of the text field widget.

        @return A widgets.TextField object.
        """
        widget = sb.widget.field.TextField(self.panel, -1, name)
        self.__addWidget(widget)
        return widget

    def createFormattedText(self):
        """Create a formatted text label widget inside the area.
        Initially the widget contains no text.

        @return A widgets.FormattedText object.
        """
        widget = sb.widget.text.FormattedText(self.panel, -1)
        self.__addWidget(widget)
        return widget

    def createRichText(self, text):
        """Create a static rich text label widget inside the area.
        Initially the widget contains no text.

        @return A widgets.FormattedText object.
        """
        widget = sb.widget.text.FormattedText(self.panel, -1, False, text)
        self.__addWidget(widget)
        return widget

    def createList(self, name, style=sb.widget.list.List.STYLE_SINGLE):
        """Create a list widget inside the area.  Initially the list
        is empty.

        @param name Identifier of the list widget.  Events sent by the
        widget are formed based on this identifier.  For example,
        selection of an item causes the notification
        '(name)-selected'.

        @return A sb.widget.list.List object.
        """
        widget = sb.widget.list.List(self.panel, self.__getNewId(), name, style)
        self.__addWidget(widget)
        return widget

    def createFormattedList(self, name):
        """Create a list widget that displays HTML-formatted items.
        Initially the list is empty.

        @param name Identifier of the list widget.  This is used when
        the widget sends notifications.

        @param A sb.widget.list.FormattedList widget.
        """
        widget = sb.widget.list.FormattedList(self.panel, self.__getNewId(), name)
        self.__addWidget(widget)
        return widget

    def createDropList(self, name):
        """Create a drop-down list widget inside the area.  Initially
        the list is empty.

        @param name Identifier of the drop-down list widget.  Events
        sent by the widget are formed based on this identifier.
        For example, selection of an item causes the notification
        '(name)-selected'.

        @return A sb.widget.list.DropList object.
        """
        widget = sb.widget.list.DropList(self.panel, self.__getNewId(), name)
        self.__addWidget(widget)
        return widget

    def createTree(self, name):
        """Create a tree widget inside the area.  Initially the tree
        is empty.  The Tree class provides methods for
        populating the tree with items.

        @param name Identifier of the tree widget.  Events
        sent by the widget are formed based on this identifier.
        For example, selection of an item causes the notification
        '(name)-selected'.

        @return A widgets.Tree object.

        @see widgets.Tree
        """
        widget = sb.widget.tree.Tree(self.panel, self.__getNewId(), name)
        self.__addWidget(widget)
        return widget

    def createTabArea(self, name, style=sb.widget.tab.TabArea.STYLE_ICON):
        """Create a TabArea widget inside the area.  The widget is
        composed of an icon list and a notebook-like tabbing area.

        @param name The identifier of the new TabArea widget.  This
        will be used as the base name of events sent by the widget.

        @return A widgets.TabArea widget.
        """
        if style == sb.widget.tab.TabArea.STYLE_FORMATTED:
            # Create a subarea and a separate MultiArea in there.
            sub = self.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)

            # Create the formatted list (with extended functionality).
            widget = sb.widget.tab.FormattedTabArea(self.panel, self.__getNewId(),
                                                    name)
            sub.setWeight(5)
            sub.setBorder(8, ui.BORDER_RIGHT)
            sub.__addWidget(widget)
            sub.setBorder(0)

            # Create the multiarea and pair it up.
            sub.setWeight(14)
            multi = sub.createMultiArea()
            widget.setMultiArea(multi)
        else:
            widget = sb.widget.tab.TabArea(self.panel, self.__getNewId(),
                                           name, style)
            self.__addWidget(widget)
        return widget

    def createCheckBox(self, name, isChecked):
        """Create a check box widget inside the area.

        @param name Identifier of the widget.  The text label of the
        widget is the translation of this identifier.

        @param isChecked The initial state of the widget.

        @return A widgets.CheckBox obejct.
        """
        widget = sb.widget.button.CheckBox(self.panel, self.__getNewId(),
                                           name, isChecked)
        self.__addWidget(widget)
        return widget

    def createRadioButton(self, name, isChecked, isFirst=False):
        widget = sb.widget.button.RadioButton(self.panel, self.__getNewId(), name,
                                              isChecked, isFirst)
        self.__addWidget(widget)
        return widget

    def createNumberField(self, name):
        """Create an integer editing widget inside the area.
        Initially the editing field is empty.

        @param name Identifier of the widget.  The text label of the
        widget is the translation of this identifier.  Notification
        events sent by the widget are based on this identifier.

        @return A widgets.NumberField object.
        """
        widget = sb.widget.field.NumberField(self.panel, self.__getNewId(), name)
        self.__addWidget(widget)
        return widget

    def createSlider(self, name):
        widget = sb.widget.bar.Slider(self.panel, self.__getNewId(), name)
        self.__addWidget(widget)
        return widget

    def createSetting(self, setting):
        self.setBorder(AREA_BORDER)
        return self.doCreateSetting(setting)

    def doCreateSetting(self, setting,
                        leftWeight=SETTING_WEIGHT_LEFT,
                        rightWeight=SETTING_WEIGHT_RIGHT):
        """Create one or more widgets that can be used to edit the
        value of the specified setting.  Automatically creates a
        subarea that contains all the widgets of the setting.

        @param setting A Setting object.

        @return  Tuple (area, widget). The area which contains all the widgets
                 of the setting, and the main widget of the setting.
        """
        if setting.getType() == 'implicit':
            # No widgets for implicit settings.
            return (None, None)

        # This is the returned main widget of the setting.
        mainWidget = None

        # Create a subarea for all the widgets of the setting.
        area = self.createArea(alignment=ui.ALIGN_HORIZONTAL)
        area.setExpanding(False)

        # All settings have a similarly positioned label on the left
        # side.
        area.setWeight(leftWeight)

        if setting.getType() != 'toggle':
            area.createText(setting.getId(), ':', 16, sb.widget.text.Text.RIGHT)
        else:
            # Check boxes use a secondary indicator label.
            label = area.createText('toggle-use-default-value',
                                    align=sb.widget.text.Text.RIGHT)

        area.setWeight(0)
        area.setBorder(2) # Space between label and setting.
        area.addSpacer()
        area.setBorder(0)

        # Create the actual widget(s) for the setting.
        area.setWeight(rightWeight)

        if setting.getType() == 'toggle':
            # Toggle settings have just a checkbox.
            isChecked = False
            if pr.getActive():
                isChecked = (pr.getActive().getValue(setting.getId()).
                             getValue() == 'yes')
            check = area.createCheckBox(setting.getId(), isChecked)
            check.setDefaultIndicator(label)
            mainWidget = check

        elif setting.getType() == 'range':
            nf = area.createNumberField(setting.getId())
            nf.setRange(setting.getMinimum(), setting.getMaximum())
            mainWidget = nf

        elif setting.getType() == 'slider':
            slider = area.createSlider(setting.getId())
            slider.setRange(setting.getMinimum(),
                            setting.getMaximum(),
                            setting.getStep())
            mainWidget = slider

        elif setting.getType() == 'choice':
            # Create a label and a drop-down list.
            drop = area.createDropList(setting.getId())
            mainWidget = drop

            # Insert the choices into the list.
            drop.addItem('default')
            for a in setting.getAlternatives():
                drop.addItem(a)

            # Sort if necessary.
            if setting.isSorted():
                drop.sortItems()

            # The selection will be updated soon afterwards.
            drop.selectItem('default')

        elif setting.getType() == 'text':
            text = area.createTextField(setting.getId())
            mainWidget = text

        elif setting.getType() == 'file':
            # Create a text field and a button.
            sub = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)
            sub.setWeight(1)
            text = sub.createTextField(setting.getId())
            mainWidget = text
            sub.setWeight(0)
            sub.setBorder(4, ui.BORDER_LEFT)
            browseButton = sub.createButton('browse-button',
                                            style=sb.widget.button.Button.STYLE_MINI)

            def browseAction():
                # Open a file browser for selecting the value for a file
                # setting.
                settingId = setting.getId()
                value = pr.getActive().getValue(settingId)
                if not value:
                    currentValue = ''
                else:
                    currentValue = value.getValue()

                selection = sb.util.dialog.chooseFile(
                    settingId + '-selection-title',
                    currentValue,
                    setting.hasToExist(),
                    setting.getAllowedTypes())

                if len(selection) > 0:
                    pr.getActive().setValue(settingId, selection)
                    if events.isMuted():
                        # Update manually.
                        text.getFromProfile(pr.getActive())

            browseButton.addReaction(browseAction)

            # If the file must exist, don't accept nonexistent files.
            if setting.hasToExist():
                text.setValidator(lambda fileName: os.path.exists(fileName))

        elif setting.getType() == 'folder':
            # Create a text field and a button, like with a file setting.
            sub = area.createArea(alignment=ui.ALIGN_HORIZONTAL, border=0)
            sub.setWeight(1)
            text = sub.createTextField(setting.getId())
            mainWidget = text
            sub.setWeight(0)
            browseButton = sub.createButton('browse-button',
                                            style=sb.widget.button.Button.STYLE_MINI)

            def folderBrowseAction():
                # Open a folder browser for selecting the value for a file
                # setting.
                settingId = setting.getId()
                value = pr.getActive().getValue(settingId)
                if not value:
                    currentValue = ''
                else:
                    currentValue = value.getValue()

                selection = sb.util.dialog.chooseFolder(settingId + '-selection-title',
                                                        currentValue)

                if len(selection) > 0:
                    pr.getActive().setValue(settingId, selection)

            browseButton.addReaction(folderBrowseAction)

            # If the file must exist, don't accept nonexistent files.
            if setting.hasToExist():
                text.setValidator(lambda fileName: os.path.exists(fileName))

        return (area, mainWidget)


class MultiArea (Area):
    """MultiArea contains multiple subareas.  Only one of them will
    be visible at a time.  All the subareas are visible in the same
    space."""

    def __init__(self, id, parentArea):
        Area.__init__(self, id, parentArea.panel, parentArea=parentArea)

        self.currentPage = None
        self.pages = {}

        events.addNotifyListener(self.onNotify, ['preparing-windows'])

    def destroy(self):
        """Destroy all the pages."""

        events.removeNotifyListener(self.onNotify)

        # Hide (detach) all the pages.
        self.showPage(None)

        for area in self.pages.values():
            area.destroy()

        Area.destroy()

    def getPages(self):
        """Return a list of all the pages of the area.

        @return An array of identifier strings.
        """
        return self.pages.keys()

    def createPage(self, identifier, align=ui.ALIGN_VERTICAL, border=3):
        """Create a subpage in the multiarea.  This does not make the
        page visible, however.

        @param identifier Identifier of the new page.
        """
        # Create a new panel for the page.
        panel = wx.Panel(self.panel, -1, style=wx.CLIP_CHILDREN)
        if host.isWindows():
            panel.SetBackgroundStyle(wx.BG_STYLE_SYSTEM)

        panel.Hide()

        # Create a new Area for the page.
        area = Area('', panel, alignment=align, border=border)
        area.setExpanding(True)
        area.setWeight(1)

        self.pages[identifier] = area

        return area

    def removePage(self, identifier):
        """Remove a page from the multiarea.

        @param identifier Page to remove.
        """
        try:
            area = self.pages[identifier]

            # Detach (if shown) and destroy the page area.
            sizer = area.panel.GetContainingSizer()
            if sizer:
                sizer.Detach(area.panel)

            area.destroy()
            area.panel.Destroy()

            # Remove the page from the dictionary.
            del self.pages[identifier]
        except KeyError:
            # Ignore unknown identifiers.
            pass

    def showPage(self, identifier):
        """Show a specific page.  The shown page is added to this
        area's containing sizer.

        @param identifier Identifier of the page to show.  If None,
        all pages are hidden.
        """
        changed = False

        for pageId in self.pages.keys():
            area = self.pages[pageId]
            sizer = area.panel.GetContainingSizer()

            if pageId == identifier:
                # Show this page.
                if not sizer:
                    self.containerSizer.Add(area.panel, 1,
                                            wx.EXPAND | wx.ALL, 3)
                    area.panel.Show()
                    changed = True
            else:
                # Hide this page, if not already hidden.
                if sizer:
                    sizer.Detach(area.panel)
                    area.panel.Hide()
                    changed = True

        if changed:
            self.updateLayout()

    def onNotify(self, event):
        if event.hasId('preparing-windows'):
            self.__updateMinSize()
            Area.updateLayout(self)

    def __updateMinSize(self):
        """Calculate min size based on all pages."""

        minWidth = 0
        minHeight = 0

        for area in self.pages.values():
            minSize = area.panel.GetBestSize()
            minWidth = max(minWidth, minSize[0])
            minHeight = max(minHeight, minSize[1])

        self.sizer.SetMinSize((minWidth + 6, minHeight + 6))
        self.sizer.Layout()

    def refresh(self):
        """Redraw the contents of the multiarea."""

        self.panel.Refresh()


class BoxedArea (Area):
    """A BoxedArea is an Area with a frame around it."""

    def __init__(self, boxTitle, parentArea, panel, alignment=0, border=0):
        """Construct a new BoxedArea.  Instances of BoxedArea are
        always sub-areas.

        @param boxTitle Identifier of the box title.

        @param parentArea The Area object inside which the BoxedArea
        will be added.

        @param panel The wxPanel into which all the widgets in the
        BoxedArea will be placed.

        @param alignment Alignment of widgets inside the area.

        @param border Border around widgets.
        """
        Area.__init__(self, '', panel, alignment, border, parentArea)

        self.label = boxTitle

        # Create a wxStaticBox and place it inside the area.
        self.box = wx.StaticBox(panel, -1, language.translate(boxTitle))

        self.containerSizer = wx.StaticBoxSizer(self.box,
                                                self.sizer.GetOrientation())

        # Insert the static box sizer into the area's sizer.  All the
        # widgets of the area will go inside the container sizer,
        # which is the static box sizer.
        self.sizer.Add(self.containerSizer, self.weight,
                       self._getLayoutFlags(), self.border)

    def destroy(self):
        """Destroy the area and its contents."""

        # First destroy the contents of the area.
        self.clear()

        self.sizer.Detach(self.containerSizer)
        ## @todo GTK crashes here if Destroy is called?
        #self.containerSizer.Destroy()

        self.box.Destroy()

        # Proceed with the normal Area destruction.
        Area.destroy(self)

    def retranslate(self):
        """Update the title of the box."""
        pass

    def createSetting(self, setting):
        self.setBorder(AREA_BORDER_BOXED)
        return self.doCreateSetting(setting,
                                    SETTING_WEIGHT_LEFT * 20 - 1,
                                    SETTING_WEIGHT_RIGHT * 20)


class ColoredArea (Area):
    """ColoredArea uses its own wxPanel so it can have a background
    color.  In other aspects it works like a regular Area."""

    def __init__(self, color, parentArea, panel, alignment=0, border=0):
        """Construct a new colored area.  Instances of ColoredArea are
        always sub-areas."""

        # The ColoredArea uses its own panel.
        colorPanel = wx.Panel(panel, -1)
        colorPanel.SetBackgroundColour(color)
        colorPanel.SetBackgroundStyle(wx.BG_STYLE_COLOUR)

        Area.__init__(self, '', colorPanel, alignment, border, parentArea)

        colorPanel.SetSizer(self.sizer)

    def getWxWidget(self):
        return self.panel
